***********************************************************************
*     file generator.F                                                *
*     adapted from COFFERGG by Markus Roth                            *
*---------------------------------------------------------------------*
*     30.01.06  Ansgar Denner     last changed  04.11.12              *
***********************************************************************





***********************************************************************
*                                                                     *
*  Representation of internal lines                                   *
*  external lines: 1,2 incoming 4,8,16, ... ougoing                   *
*  internal lines: sum of external lines that contribute              *
*                  to momentum of internal lines                      *
*  allbinary = sum of all external lines = 1+2+4+8+16+32+ ...         *
*                                                                     *
*  Representation of diagrams/channels                                *
*                                                                     *
*  channels are stored as follows:                                    *
*          ninv(cha,gen)=next-4      number of invariants             *
*  incoming invariants for decays:                                    *
*          ininv(ns,cha,gen)         incoming particle                *
*          idhepinv(ns,cha,gen)      idhep of incoming particle       *
*          powerinv(nt,cha,gen)      power of incoming particle       *
*  outgoing invariant for t-channel processes:                        *
*          ininv(maxnt-nt,cha,gen)   outgoing particle                *
*          idhepinv(ns,cha,gen)      idhep of outcoming particle =0   *
*          powerinv(nt,cha,gen)      power of outcoming particle =0   *
*  decays: indecay(maxns-ns,cha,gen) incoming particle                *
*          out1decay(maxns-ns,cha,gen) outcoming particle 1           *
*          out2decay(maxns-ns,cha,gen) outcoming particle 2           *
*  t-channel processes:                                               *
*          in1process(nt,cha,gen)    incoming particle 1              *
*          in2process(nt,cha,gen)    incoming particle 2              *
*          out1process(nt,cha,gen)   outgoing particle 1              *
*          out2process(nt,cha,gen)   outgoing particle 2              *
*          inprocess(nt,cha,gen)     all incoming particles           *
*          virtprocess(nt,cha,gen)   virtual particle                 *
*          idhepprocess(nt,cha,gen)  idhep of virtual particle        *
*          powerprocess(nt,cha,gen)  power of virtual particle        *
*                                                                     *
***********************************************************************

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc****
c                                                                     c
c     Monte Carlo generator                                           c
c                                                                     c
c     written by Markus Roth                                          c
c                                                                     c
*     last changed 01.12.2006  Ansgar Denner                          *
*                                                                     *
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc****
***********************************************************************
      subroutine generate_momenta(random,kbeam,k,kcms,
     &    g,channel,generator,switch)
***********************************************************************
*     generation of momenta                                           *
*     input:                                                          *
*       random:    random numbers                                     *
*       kbeam:     beam momenta                                       *
*       channel:   channel                                            *
*       generator: generator                                          *
*     output:  for switch = 0                                         *
*       k:       generated momenta in lab system                      *
*       kcms:    generated momenta in cms system                      *
*       switch > 0 for error                                          *
*     irrelevant:                                                     *
*       g:       density                                              *
*     adapted from COFFERGG by Markus Roth                            *
*---------------------------------------------------------------------*
*     Ansgar Denner 16.03.07   last changed  16.03.07   Ansgar Denner *
***********************************************************************
      implicit none

#include "mcparams.h"
#include "mccommon.h"

c local variables
      real*8 random(3*maxe-7)
      real*8 k(maxe,0:3),kcms(maxe,0:3),kbeam(2,0:3),g
      real*8 kbeamcms(2,0:3)
      real*8 gamma,dir,ebeamcms,dirbeamcms
      integer channel,generator,switch
c      logical lcmsgen

      real*8 ktest(maxe,0:3),q(0:3),q1(0:3),sij(maxe,maxe)


c      write(*,*) 'generate_momenta: generator= ',generator,
c     &    nexternal(generator)

c      lcmsgen=.false.
c      lcmsgen=.true.

      gamma=(kbeam(1,0)+kbeam(2,0))/(2d0*sqrt(kbeam(1,0)*kbeam(2,0)))
      dir=-dsign(1d0,kbeam(1,3)+kbeam(2,3))

      if(lcmsgen) then
        ebeamcms=sqrt(kbeam(1,0)*kbeam(2,0))
        dirbeamcms=dsign(1d0,kbeam(1,3)-kbeam(2,3))
        kbeamcms(1,0)=ebeamcms
        kbeamcms(2,0)=ebeamcms
        kbeamcms(1,3)=ebeamcms*dirbeamcms
        kbeamcms(2,3)=-ebeamcms*dirbeamcms
        call generation(random,kbeamcms,kcms,g,channel,generator,
     *      switch)
        call boostz(gamma,-dir,kcms,k,nexternal(generator))


      else
        call generation(random,kbeam,k,g,channel,generator,
     &      switch)
        call boostz(gamma,dir,k,kcms,nexternal(generator))






      endif


      end

***********************************************************************
      subroutine boostz(gamma,dir,k,kboosted,lnext)
***********************************************************************
*     boost momentum set along z-axis                                 *
*     dir = +1: boosts k along positve z axis                         *
*     dir = -1: boosts k along negative z axis                        *
*     input                                                           *
*     gamma     Lorent factor                                         *
*     dir       sign of beta                                          *
*     k         input momentum                                        *
*     input                                                           *
*     kcms      boosted momentum                                      *
*---------------------------------------------------------------------*
*     Ansgar Denner 16.03.07    last changed  16.03.07 Ansgar Denner  *
***********************************************************************
      implicit none

#include "mcparams.h"
#include "mccommon.h"

c local variable
      real*8 k(maxe,0:3),kboosted(maxe,0:3),dir
      real*8 beta,gamma
      integer lnext

      beta=dir*sqrt((1d0-1d0/gamma)*(1d0+1d0/gamma))
      do ext=1,lnext
        kboosted(ext,0)=gamma*(k(ext,0)+beta*k(ext,3))
        kboosted(ext,1)=k(ext,1)
        kboosted(ext,2)=k(ext,2)
        kboosted(ext,3)=gamma*(k(ext,3)+beta*k(ext,0))
      enddo
      end

***********************************************************************
      subroutine generation(random,kbeam,k,g,channel,generator,
     *  switch)
***********************************************************************
*     generation of momenta                                           *
*     input:                                                          *
*       random:    random numbers                                     *
*       kbeam:     beam momenta                                       *
*       channel:   channel                                            *
*       generator: generator                                          *
*     output:  for switch = 0                                         *
*       k:       generated momenta                                    *
*       switch > 0 for error                                          *
*     irrelevant:                                                     *
*       g:       density                                              *
*     adapted from COFFERGG by Markus Roth                            *
*---------------------------------------------------------------------*
*     by Markus Roth         last changed   14.11.06   Ansgar Denner  *
***********************************************************************
      implicit none

#include "mcparams.h"
#include "mccommon.h"

c local variables
      real*8 random(3*maxe-7),p(0:3,2**maxe),s(2**maxe)
      real*8 k(maxe,0:3),kt(maxe,0:3),kbeam(2,0:3),mapx
      real*8 mmin,mmax,smin,smax,x,g,tcut1,tcut2,sum
      integer i1,i2,i3,i4,ns,nt,channel,generator,switch,step
      integer ranstart,virtinv,inv1,inv2,em,sp,co



c use of random numbers
c random 1 -- ninv:                      invariants
c random (ninv+1) -- (ninv+2*nprocess):  angles for t-channel processes
c random (ninv+2*nprocess+1) -- (ninv+2*nprocess+2*ndecay):
c                                        angles for s-channel decays
c random ninv+2*nprocess+2*ndecay+1:     energy fraction after photon radiation
c
c n-particle phase space:
c    n-2 invariants +
c    2(n-1) angles in (n-1) decays and t-channel processes
c => 3(n-1)-1 variables
c    + energy fraction after photon radiation
c => 3(n-1) = 3(next-3) variables for Born
c    3n-4   = 3next-10 variables for bremsstrahlung

c      write(*,*) 'generation: random =  ',random

      sum=0d0
      if(switch.gt.0)return
c      step=2 ! calculation of local density
      step=1 ! calculation of momenta
      em=emmap(channel,generator)
      sp=spmap(channel,generator)
      co=comap(channel,generator)

c      write(*,*) 'generation: em',em,sp,co,generator,channel
c      write(*,*) 'generation: channel',channel

      if(step.eq.1)then
c incoming momenta and definition of x
        ranstart=1+ninv(channel,generator)
     *    +2*nprocess(channel,generator)+2*ndecay(channel,generator)

c        write(*,*) 'generation: use random =  ',ranstart,
c     &      random(ranstart)

        if(em.eq.1.or.(sp.eq.1.and.em.ge.3))then
c changed 17.11.06
c          x=1d0-mapinv(random(ranstart),powerexmap,0d0,0d0,1d0,switch)
          x=1d0-mapx(random(ranstart),powerexmap,0d0,0d0,1d0-xminp,
     &         switch)
          do i1=0,3
            p(i1,1)=-x*kbeam(1,i1)
            p(i1,2)=-kbeam(2,i1)
          enddo
        elseif(em.eq.2.or.(sp.eq.2.and.em.ge.3))then
c changed 17.11.06
c          x=1d0-mapinv(random(ranstart),powerexmap,0d0,0d0,1d0,switch)
          x=1d0-mapx(random(ranstart),powerexmap,0d0,0d0,1d0-xminp,
     &      switch)
          do i1=0,3
            p(i1,1)=-kbeam(1,i1)
            p(i1,2)=-x*kbeam(2,i1)
          enddo
        else
          x=1d0
          do i1=0,3
            p(i1,1)=-kbeam(1,i1)
            p(i1,2)=-kbeam(2,i1)
          enddo
        endif
c remaining momenta
        do i1=0,3
          p(i1,3)=p(i1,1)+p(i1,2)
          p(i1,allbinary(co,generator)-1)=-p(i1,1)
          p(i1,allbinary(co,generator)-2)=-p(i1,2)
          p(i1,allbinary(co,generator)-3)=-p(i1,3)
        enddo
c square of center-of-mass energy
c        write(nout,*) ' generation: kb1 =',(kbeam(1,i1),i1=0,3)
c        write(nout,*) ' generation: kb2 =',(kbeam(2,i1),i1=0,3)
c        write(nout,*) ' generation: p1  =',(p(i1,1),i1=0,3)
c        write(nout,*) ' generation: p2  =',(p(i1,2),i1=0,3)
c        write(nout,*) ' generation: p3  =',(p(i1,3),i1=0,3)
        s(3)=p(0,3)**2-p(1,3)**2-p(2,3)**2-p(3,3)**2
        s(allbinary(co,generator)-3)=s(3)
c event excluded by cuts after intial-state radiation
        if(s(3).le.mcutinv(allbinary(co,generator)-3,generator)**2)then
          switch=1
          if(numout.lt.maxout.and.lnoutmc.ge.5)then
c            write(nout,'(" event = ",i9,",",(a))') nevt,
 100        format(/" event = ",i9,", gen = ",i4,", channel = ",i4,
     &          ", exit = ",i4,(a))
            write(nout,100) nevt,cogen,cocha,switch,
     &          ', generation: s < sppmin'
            write(nout,*) ' s,sppmin =',s(3),
     &          mcutinv(allbinary(co,generator)-3,generator)**2
            numout=numout+1
          endif
          return
        endif
      else
c calculation of local density
        g=1d0
c defining momenta
        do i1=1,nexternal(generator)
        do i2=0,3
          kt(i1,i2)=k(i1,i2)
        enddo
        enddo
* map momenta from n to (n-1) particle phase space
        call mapin(kt,g,em,sp,co,nexternal(generator),switch)
        do i1=1,allbinary(co,generator)
          p(0,i1)=0d0
        enddo
        do i1=0,3
          p(i1,1)=-kt(1,i1)
          p(i1,2)=-kt(2,i1)
          do i2=3,nexternal(generator)
            p(i1,2**(i2-1))=kt(i2,i1)
          enddo
        enddo
c calculating invariants
        do i1=1,allbinary(co,generator)-1
          if(p(0,allbinary(co,generator)-i1).eq.0d0)then
            do i2=1,nexternal(generator)
              i3=i1/2**(i2-1)
              if(2*(i3/2).ne.i3.and.i1.ne.2**(i2-1))then
                do i4=0,3
                  p(i4,i1)=p(i4,i1-2**(i2-1))+p(i4,2**(i2-1))
                enddo
                goto 900
              endif
            enddo
 900        s(i1)=p(0,i1)**2-p(1,i1)**2-p(2,i1)**2-p(3,i1)**2
          else
            do i2=0,3
              p(i2,i1)=-p(i2,allbinary(co,generator)-i1)
            enddo
            s(i1)=s(allbinary(co,generator)-i1)
          endif
        enddo
      endif




c external masses
      do i1=1,nexternal(generator)
      if(i1.ne.co)then
        s(2**(i1-1))=massext2(i1,generator)
        s(allbinary(co,generator)-2**(i1-1))=massext2(i1,generator)
      endif
      enddo
c inv
      do ns=1,ninv(channel,generator)
        inv1=ininv(ns,channel,generator)
        inv2=allbinary(co,generator)-3-ininv(ns,channel,generator)
        mmin=mcutinv(inv1,generator)
        mmax=mcutinv(inv2,generator)
        do i1=1,ns-1
          virtinv=ininv(i1,channel,generator)
          if(lmin(i1,ns,channel,generator).and.
     *      sqrt(s(virtinv)).gt.mcutinv(virtinv,generator))then
            mmin=mmin+sqrt(s(virtinv))-mcutinv(inv1,generator)
     *        +mcutinv(inv1-virtinv,generator)
            inv1=inv1-virtinv
          endif
          if(lmax(i1,ns,channel,generator).and.
     *      sqrt(s(virtinv)).gt.mcutinv(virtinv,generator))then
            mmax=mmax+sqrt(s(virtinv))-mcutinv(inv2,generator)
     *        +mcutinv(inv2-virtinv,generator)
            inv2=inv2-virtinv
          endif
        enddo
        smin=mmin**2
        smax=(sqrt(s(3))-mmax)**2


c        write(*,*) 'integrand use random(ns)',ns,random(ns)

        call inv(
     *    random(ns),                                    ! random number
     *    s(ininv(ns,channel,generator)),                ! invariant mass
     *    g,                                             ! local density
     *    mcmass(idhepinv(ns,channel,generator)),        ! mass
     *    mcwidth(idhepinv(ns,channel,generator)),       ! width
     *    powerinv(ns,channel,generator),                ! mapping
     *    smin,                                          ! lower bound
     *    smax,                                          ! upper bound
     *    step,switch)
c        write(*,*) 'generation afterinv'
      enddo
c process
      ranstart=ninv(channel,generator)
      do nt=1,nprocess(channel,generator)
        tcut1=0d0
        tcut2=0d0
        if(nt.eq.1)then
          tcut1=tcutprocess(allbinary(co,generator)
     *      -virtprocess(nt,channel,generator),generator)
          tcut2=tcutprocess(virtprocess(nt,channel,generator),
     *      generator)
        endif

c        write(*,*) 'integrand use random(nt)',ranstart+2*nt-1,
c     &      random(ranstart+2*nt-1),random(ranstart+2*nt)

        call process(
     *    random(ranstart+2*nt-1),                       ! random numbers
     *    p(0,in1process(nt,channel,generator)),         ! incomming particle 1
     *    p(0,in2process(nt,channel,generator)),         ! incomming particle 2
     *    p(0,out1process(nt,channel,generator)),        ! outgoing particle 1
     *    p(0,out2process(nt,channel,generator)),        ! outgoing particle 2
     *    p(0,virtprocess(nt,channel,generator)),        ! virtual particle
     *    g,                                             ! local density
     *    mcmass(idhepprocess(nt,channel,generator)),    ! mass
     *    mcwidth(idhepprocess(nt,channel,generator)),   ! width
     *    powerprocess(nt,channel,generator),            ! mapping
     *    s(inprocess(nt,channel,generator)),            ! incomming particles
     *    s(out1process(nt,channel,generator)),          ! outgoing particle 1
     *    s(out2process(nt,channel,generator)),          ! outgoing partcile 2
     *    s(virtprocess(nt,channel,generator)),          ! virtual particle
     *    s(in1process(nt,channel,generator)),           ! incomming particle 1
     *    s(in2process(nt,channel,generator)),           ! incomming particle 2
     *    tcut1,tcut2,                                   ! angular cuts
     *    step,switch)
      enddo
c decay
      ranstart=ninv(channel,generator)+2*nprocess(channel,generator)
      do ns=1,ndecay(channel,generator)

c        write(*,*) 'integrand use random(nd)',ranstart+2*ns-1,
c     &      random(ranstart+2*ns-1),random(ranstart+2*ns)

        call decay(
     *    random(ranstart+2*ns-1),                       ! random numbers
     *    p(0,indecay(ns,channel,generator)),            ! incomming particle
     *    p(0,out1decay(ns,channel,generator)),          ! outgoing particle 1
     *    p(0,out2decay(ns,channel,generator)),          ! outgoing particle 2
     *    g,                                             ! local density
     *    s(indecay(ns,channel,generator)),              ! incomming particle
     *    s(out1decay(ns,channel,generator)),            ! outgoing particle 1
     *    s(out2decay(ns,channel,generator)),            ! outgoing particle 2
     *    step,switch)
      enddo

c defining external momenta
      do i1=0,3
        k(1,i1)=-p(i1,1)
        k(2,i1)=-p(i1,2)
        do i2=3,nexternal(generator)
          k(i2,i1)=p(i1,2**(i2-1))
        enddo
      enddo
c map
      if(step.eq.1)then
        ranstart=1+ninv(channel,generator)
     *    +2*nprocess(channel,generator)+2*ndecay(channel,generator)
        call mapout(random(ranstart),k,x,g,em,sp,co,
     *    nexternal(generator),switch)
      endif
c restoring external masses
      do i1=1,nexternal(generator)
        k(i1,0)=sqrt(massext2(i1,generator)
     *    +k(i1,1)**2+k(i1,2)**2+k(i1,3)**2)
        if(i1.eq.2) then
          sum=k(1,0)+k(2,0)
        else
          sum=sum-k(i1,0)
        endif
      enddo
c      if(dabs(sum).gt.1d-9*dabs(k(1,0)+k(2,0))) then
c        write(*,*) 'generation: momentum conservation violated',
c     &      sum,1d-9*dabs(k(1,0)+k(2,0))
c        write(*,105) (i2,0,k(i2,0),i2=1,nexternal(generator))
c      endif
c      if(.not.(sum.eq.0d0.or.sum.ne.0d0)) then
c        write(*,*) 'generation: momentum conservation violated NaN'
c     &      ,sum,1d-9*dabs(k(1,0)+k(2,0))
c        write(*,105) (i2,0,k(i2,0),i2=1,nexternal(generator))
c      endif

      call checkmom(k,nexternal(generator),generator,switch)




      end

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                 c
c     calculating densities                                       c
c                                                                 c
c     written by Markus Roth                                      c
c                                                                 c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine checkmom(k,lnext,generator,switch)
***********************************************************************
*     checks momentum conservation and mass shell conditions          *
*     input:                                                          *
*       random:    random numbers                                     *
*       k:         external particle momenta                          *
*       lnext:     number of external particles                       *
*       generator: generator                                          *
*     output:                                                         *
*       switch > 0 for error                                          *
*     adapted from COFFERGG by Markus Roth                            *
*---------------------------------------------------------------------*
*     by Markus Roth         last changed   05.01.07   Ansgar Denner  *
***********************************************************************
      implicit none

#include "mcparams.h"
#include "mccommon.h"

c local variables
      real*8 k(maxe,0:3),mom,acc
      integer i1,i2,lnext,switch,generator

      parameter(acc=1d-9)

      if(switch.gt.0)return
      do i1=0,3
        mom=k(1,i1)+k(2,i1)
        do i2=3,lnext
          mom=mom-k(i2,i1)
        enddo
        if(dabs(mom).gt.acc*dabs(k(1,0)+k(2,0)))then
          switch=35
          if(numout.lt.maxout.and.lnoutmc.ge.5)then
            numout=numout+1
 120        format(/" event = ",i9,", gen = ",i4,", channel = ",i4,
     &          ", exit = ",i4,
     &          ", momentum conservation violated:",i2,2d24.16)
            write(nout,120) nevt,cogen,cocha,switch,
     &          i1,mom,acc*dabs(k(1,0)+k(2,0))

c            write(nout,'(" event= ",i9,",",
c     &          " momentum conservation violated:",i2,2d24.16)')
c     *          nevt,i1,mom,acc*dabs(k(1,0)+k(2,0))
            write(nout,200) (i2,i1,k(i2,i1),i2=1,lnext)
 200        format(6x,' k(',i1,',',i1,')    = ',d20.14)
          endif
          return
        endif
      enddo
      do i1=1,lnext
        mom=k(i1,0)**2-massext2(i1,generator)
        do i2=3,1,-1
          mom=mom-k(i1,i2)**2
        enddo
        if(dabs(mom).gt.acc*dabs(k(1,0)+k(2,0))**2)then
          switch=36
          if(numout.lt.maxout.and.lnoutmc.ge.5)then
            numout=numout+1
 110        format(/" event = ",i9,", gen = ",i4,", channel = ",i4,
     &          ", exit = ",i4,
     &          ", onshellness violated:",i2,2d24.16)
            write(nout,110) nevt,cogen,cocha,switch,
     &          i1,mom,acc*dabs(k(1,0)+k(2,0))**2
c            write(nout,'(" event= ",i9,",",
c     &          " onshellness violated:",i2,2d24.16)')
c     &          nevt,i1,mom,
c     &          acc*dabs(k(1,0)+k(2,0))**2
            write(nout,'(" mass2 =",6x,1d24.16)') massext2(i1,generator)
            write(nout,200) (i1,i2,k(i1,i2),i2=0,3)
          endif
          return
        endif
      enddo
      end

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                 c
c     calculating densities                                       c
c                                                                 c
c     written by Markus Roth                                      c
c                                                                 c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
*                                                                 *
*     calculates densities for all channels in one go!            *
*     owing to appropriate initialization of invariants           *
*     generation can do this only for each channel separately     *
*                                                                 *
*******************************************************************
      subroutine density(k,g,generator,switch)
***********************************************************************
*     calculation of phase space density                              *
*     input:                                                          *
*       k:       generated momenta                                    *
*       generator: generator                                          *
*     output:  for switch = 0                                         *
*       g:       density                                              *
*       switch > 0 for error                                          *
*     adapted from COFFERGG by Markus Roth                            *
*---------------------------------------------------------------------*
*     by Markus Roth         last changed   14.11.06   Ansgar Denner  *
***********************************************************************
      implicit none

#include "mcparams.h"
#include "mccommon.h"

c local variables
      real*8 random(3*maxe-7),p(0:3,2**maxe,0:maxe,0:maxe,0:maxe)
      real*8 s(2**maxe,0:maxe,0:maxe,0:maxe),k(maxe,0:3),kt(maxe,0:3)
      real*8 gmap(0:maxe,0:maxe,0:maxe),ginv(maxiv),gprocess(maxpr)
      real*8 gdecay(maxdc),g(maxch),mmin,mmax,smin,smax,tcut1,tcut2
      real*8 sinv,sum
      integer i1,i2,i3,i4,ns,nt,channel,generator,switch,step
      integer virtinv,inv1,inv2,em,sp,co

      if(switch.gt.0)return


      step=2
      do em=0,nexternal(generator)
      do sp=0,nexternal(generator)
      do co=0,nexternal(generator)
c        write(*,*) 'density, lmap',
c     &      em,sp,co,generator,lmap(em,sp,co,generator)
        if(lmap(em,sp,co,generator))then
          do i1=1,nexternal(generator)
          do i2=0,3
            kt(i1,i2)=k(i1,i2)
          enddo
          enddo
c      if(.not.(sum.eq.0d0.or.sum.ne.0d0)) then
c        write(*,*) 'density: momentum conservation violated NaN'
c     &      ,sum,1d-9*dabs(k(1,0)+k(2,0))
c        write(*,105) (i2,0,k(i2,0),i2=1,nexternal(generator))
c      endif
c          call checkmom(k,nexternal(generator),generator,switch)
c generating phase space configurations
          call mapin(kt,gmap(em,sp,co),em,sp,co,nexternal(generator),
     *      switch)
c            write(*,*) 'density: mapin switch= ',switch
          if(co.gt.0)
     &        call checkmom(kt,nexternal(generator)-1,generator,switch)
c      if(.not.(sum.eq.0d0.or.sum.ne.0d0)) then
c        write(*,*) 'density: co momentum conservation violated NaN'
c     &      ,sum,1d-9*dabs(k(1,0)+k(2,0))
c        write(*,105) (i2,0,k(i2,0),i2=1,nexternal(generator))
c      endif
          if(switch.gt.0) return
c calculating invariants
          do i1=1,allbinary(co,generator)
            p(0,i1,em,sp,co)=0d0
          enddo
          do i1=0,3
            p(i1,1,em,sp,co)=-kt(1,i1)
            p(i1,2,em,sp,co)=-kt(2,i1)
            do i2=3,nexternal(generator)
              p(i1,2**(i2-1),em,sp,co)=kt(i2,i1)
            enddo
          enddo

c          write(*,*) 'density allbin ',co,generator,
c     &        allbinary(co,generator)
c calculating invariants
          do i1=1,allbinary(co,generator)-1

c            if(i1.eq.3) write(*,*) 'density test',
c     &          p(0,allbinary(co,generator)-i1,em,sp,co)

            if(p(0,allbinary(co,generator)-i1,em,sp,co).eq.0d0)then
              do i2=1,nexternal(generator)
                i3=i1/2**(i2-1)
                if(2*(i3/2).ne.i3.and.i1.ne.2**(i2-1))then
                  do i4=0,3
                    p(i4,i1,em,sp,co)=p(i4,i1-2**(i2-1),em,sp,co)
     *                +p(i4,2**(i2-1),em,sp,co)
                  enddo
                  goto 900
                endif
              enddo
 900          s(i1,em,sp,co)=p(0,i1,em,sp,co)**2-p(1,i1,em,sp,co)**2
     *          -p(2,i1,em,sp,co)**2-p(3,i1,em,sp,co)**2

c            if(i1.eq.3) write(*,*) 'density test a ',
c     &            em,sp,co,s(i1,em,sp,co)

            else
              do i2=0,3
                p(i2,i1,em,sp,co)=
     *            -p(i2,allbinary(co,generator)-i1,em,sp,co)
              enddo
              s(i1,em,sp,co)=s(allbinary(co,generator)-i1,em,sp,co)
c            if(i1.eq.3) write(*,*) 'density test b ',s(i1,em,sp,co)
            endif
          enddo
c external masses
          do i1=1,nexternal(generator)
          if(i1.ne.co)then
            s(2**(i1-1),em,sp,co)=massext2(i1,generator)
            s(allbinary(co,generator)-2**(i1-1),em,sp,co)=
     *        massext2(i1,generator)
          endif
          enddo
        endif
      enddo
      enddo
      enddo

c inv
      do i1=1,maxinv(generator)
        ginv(i1)=1d0
        channel=chinv(i1,generator)
        em=emmap(channel,generator)
        sp=spmap(channel,generator)
        co=comap(channel,generator)
        ns=nsinv(i1,generator)
        inv1=ininv(ns,channel,generator)
        inv2=allbinary(co,generator)-3-ininv(ns,channel,generator)
        mmin=mcutinv(inv1,generator)
        mmax=mcutinv(inv2,generator)
        do i2=1,ns-1
          virtinv=ininv(i2,channel,generator)
          sinv=s(virtinv,em,sp,co)
          if(sinv.lt.0d0)then
            if(sinv.gt.techparam(2))then
              sinv=0d0
            else
              if(numout.lt.maxout.and.lnoutmc.ge.5)then
 100        format(/" event = ",i9,", gen = ",i4,", channel = ",i4,
     &              ", exit = 20",(a))
            write(nout,100) nevt,cocha,1,
c                write(nout,'(" event = ",i9,",",(a))') nevt,
     &              ', density: sinv < 0'
                numout=numout+1
              endif
              switch=20
              return
            endif
          endif
          if(lmin(i2,ns,channel,generator).and.
     *      sqrt(sinv).gt.
     *      mcutinv(virtinv,generator))then
            mmin=mmin+sqrt(sinv)
     *        -mcutinv(inv1,generator)
     *        +mcutinv(inv1-virtinv,generator)
            inv1=inv1-virtinv
          endif
          if(lmax(i2,ns,channel,generator).and.
     *      sqrt(sinv).gt.
     *      mcutinv(virtinv,generator))then
            mmax=mmax+sqrt(sinv)
     *        -mcutinv(inv2,generator)
     *        +mcutinv(inv2-virtinv,generator)
            inv2=inv2-virtinv
          endif
        enddo
        smin=mmin**2
        smax=(sqrt(s(3,em,sp,co))-mmax)**2


        call inv(
     *    random(1),                                        ! random number
     *    s(ininv(ns,channel,generator),em,sp,co),          ! invariant mass
     *    ginv(i1),                                         ! local density
     *    mcmass(idhepinv(ns,channel,generator)),           ! mass
     *    mcwidth(idhepinv(ns,channel,generator)),          ! width
     *    powerinv(ns,channel,generator),                   ! mapping
     *    smin,                                             ! lower bound
     *    smax,                                             ! upper bound
     *    step,switch)
c         write(*,*) 'density afterinv',
c     &      ininv(ns,channel,generator),em,sp,co,
c     &      s(ininv(ns,channel,generator),em,sp,co)
      enddo
c      write(*,*) 'density: inv switch= ',switch

c process
      do i1=1,maxprocess(generator)
        gprocess(i1)=1d0
        channel=chprocess(i1,generator)
        em=emmap(channel,generator)
        sp=spmap(channel,generator)
        co=comap(channel,generator)
        nt=ntprocess(i1,generator)
        tcut1=0d0
        tcut2=0d0
        if(nt.eq.1)then
          tcut1=tcutprocess(allbinary(co,generator)
     *      -virtprocess(nt,channel,generator),generator)
          tcut2=tcutprocess(virtprocess(nt,channel,generator),
     *      generator)
        endif
c        write(*,*) 'density befproc',
c     &      inprocess(nt,channel,generator),
c     &      out1process(nt,channel,generator),
c     &      out2process(nt,channel,generator)
        call process(
     *    random,                                           ! random numbers
     *    p(0,in1process(nt,channel,generator),em,sp,co),   ! incomming p. 1
     *    p(0,in2process(nt,channel,generator),em,sp,co),   ! incomming p. 2
     *    p(0,out1process(nt,channel,generator),em,sp,co),  ! outgoing p. 1
     *    p(0,out2process(nt,channel,generator),em,sp,co),  ! outgoing p. 2
     *    p(0,virtprocess(nt,channel,generator),em,sp,co),  ! virtual p.
     *    gprocess(i1),                                     ! local density
     *    mcmass(idhepprocess(nt,channel,generator)),       ! mass
     *    mcwidth(idhepprocess(nt,channel,generator)),      ! width
     *    powerprocess(nt,channel,generator),               ! mapping
     *    s(inprocess(nt,channel,generator),em,sp,co),      ! incomming p.s
     *    s(out1process(nt,channel,generator),em,sp,co),    ! outgoing p. 1
     *    s(out2process(nt,channel,generator),em,sp,co),    ! outgoing p. 2
     *    s(virtprocess(nt,channel,generator),em,sp,co),    ! virtual p.
     *    s(in1process(nt,channel,generator),em,sp,co),     ! incomming p. 1
     *    s(in2process(nt,channel,generator),em,sp,co),     ! incomming p. 2
     *    tcut1,tcut2,                                      ! angular cuts
     *    step,switch)
c         write(*,*) 'density aftproc'
      enddo
c decay
      do i1=1,maxdecay(generator)
        gdecay(i1)=1d0
        channel=chdecay(i1,generator)
        em=emmap(channel,generator)
        sp=spmap(channel,generator)
        co=comap(channel,generator)
        ns=nsdecay(i1,generator)
        call decay(
     *    random,                                           ! random numbers
     *    p(0,indecay(ns,channel,generator),em,sp,co),      ! incomming p.
     *    p(0,out1decay(ns,channel,generator),em,sp,co),    ! outgoing p. 1
     *    p(0,out2decay(ns,channel,generator),em,sp,co),    ! outgoing p. 2
     *    gdecay(i1),                                       ! local density
     *    s(indecay(ns,channel,generator),em,sp,co),        ! incomming p.
     *    s(out1decay(ns,channel,generator),em,sp,co),      ! outgoing p. 1
     *    s(out2decay(ns,channel,generator),em,sp,co),      ! outgoing p. 2
     *    step,switch)
       enddo
c calculating densities
      do channel=1,nchannel(generator)
c        write(*,*) 'density channel ',channel,generator
        em=emmap(channel,generator)
        sp=spmap(channel,generator)
        co=comap(channel,generator)
        g(channel)=gmap(em,sp,co)
c inv
        do ns=1,ninv(channel,generator)
          g(channel)=g(channel)
     *      *ginv(numinv(ns,channel,generator))
c          write(*,*) 'density ginv', ns,g(channel),
c     &        ginv(numinv(ns,channel,generator))
        enddo
c process
        do nt=1,nprocess(channel,generator)
          g(channel)=g(channel)
     *      *gprocess(numprocess(nt,channel,generator))
c          write(*,*) 'density gpro', nt,g(channel),
c     &       gprocess(numprocess(nt,channel,generator))
        enddo
c decay
        do ns=1,ndecay(channel,generator)
          g(channel)=g(channel)
     *      *gdecay(numdecay(ns,channel,generator))
c          write(*,*) 'density gdec', ns,g(channel),
c     &       gdecay(numdecay(ns,channel,generator))
        enddo
      enddo



      end

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                 c
c     initialization generator                                    c
c                                                                 c
c     written by Markus Roth                                      c
c                                                                 c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
***********************************************************************
      subroutine initgenerator(lenergy,sppmin,lhepnum,generator,lnext,
     *  lsmodel,lsincludecuts,lssub,flag)
***********************************************************************
*     initialization of phase space generator                         *
*     adapted from COFFERGG by Markus Roth                            *
*     input:  lenergy,lhepnum,generator,lnext                         *
*             lsmodel,lsincludecuts,lssub                             *
*     output: sppmin,flag                                             *
*---------------------------------------------------------------------*
*     by Markus Roth         last changed   22.02.10   Ansgar Denner  *
***********************************************************************
      implicit none

#include "mcparams.h"
#include "mccommon.h"

c local variables
      real*8 m2,m3,e2,e3,scutinv,lenergy,sppmin
      integer idhep(maxe,maxe),binary(maxe,maxe)
      integer i1,i2,i3,ns,nt,maxns,maxnt,binary1,binary2
      integer binary3,in1(maxe),in2(maxe),out1(maxe),out2(maxe),pid
c      integer virt(maxe),channel,generator,lsincludecuts,prop2,prop3
      integer virt(maxe),channel,generator,lsincludecuts
      integer smap,lsmodel,lhepnum(maxe),lnext
      integer em,sp,co,n,lssub
      integer flag,lastoutline
c      logical vertex,included,exist,comparechannel,compareinv
      logical vertex,included,comparechannel,compareinv
      logical comparedecay,compareprocess,schannel
c      character*9 particle(maxv)
c      character*80 vertices


c      write(*,*) 'initgenerator: input', lenergy,sppmin,lhepnum,
c     &    generator,lnext,
c     *    lsmodel,lsincludecuts,lssub,flag

      if(generator.gt.maxg)then
        write(*,'(a28,i6)')' initgenerator: reset maxg >',
     *      generator
        stop
      endif


      flag = 0

c options
c     lnoutgen=1                       prints summary of channels
c     lnoutgen=2                       prints all channels
      smap=1                         ! mapping
c      smap=0                         ! mapping
c technical parameter in h function and subroutine process
c      techparam(1)=1d-4              ! small necotive mass in h-function
c      techparam(1)=1d-8              ! small necotive mass in h-function
c      techparam(2)=1d-10*lenergy**2  ! allowed uncertainty tmax in process
c      techparam(3)=1d-12*lenergy     ! allowed uncertainty q1,q2 in process
c      powerexmap=0.8d0                 ! mapping of x in radiative events
c number of external particles
      nexternal(generator)=lnext
c      write(*,*) 'nexternal',nexternal(generator),lnext

c initialize binary counting of particle number
      do ext=1,maxe
      do ext2=1,maxe
        binary(ext,ext2)=0
      enddo
      enddo
      do ext=1,nexternal(generator)
        binary(ext,1)=2**(ext-1)
      enddo
      allbinary(0,generator)=0
      do ext=1,nexternal(generator)
        allbinary(0,generator)=allbinary(0,generator)+2**(ext-1)
      enddo
      do co=1,nexternal(generator)
        allbinary(co,generator)=allbinary(0,generator)-2**(co-1)
      enddo

c      write(*,*) 'initgenerator: allbinary',generator,
c&      (allbinary(i1,generator),i1=0,nexternal(generator))

      call init_mcmodel(smap)

c particle identity
* idhep(next,1) = hepnumber
      do ext=1,nexternal(generator)
        idhep(ext,1)=lhepnum(ext)
        if(ext.ge.3)idhep(ext,1)=-idhep(ext,1)
        massext2(ext,generator)=mcmass(abs(lhepnum(ext)))**2
c        write(*,*) 'initgenerator: massext2: ',ext,generator,
c     &      massext2(ext,generator)
      enddo

c uniform sampling for checks (after massext2 definition!)
c different small masses are needed to preserve structure of
c channels
      if(smap.eq.0)then
        do i1=1,naux-1
          if(mcmass(i1).ne.0d0)mcmass(i1)=1d-5*i1
          mcwidth(i1)=0d0
          power(i1)=1d-5*i1
        enddo

        if(generator.gt.0) then
          do i1=1,nexternal(0)
            massext2(i1,generator)=massext2(i1,0)
          enddo
        endif
      endif

c small fermion-higgs couplings
      do i1=0,maxv
        light(i1)=0
        if(mcmass(i1).eq.0d0)light(i1)=1
        light(-i1)=light(i1)
      enddo

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                 c
c     initialize cuts                                             c
c                                                                 c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      do i1=0,allbinary(0,generator)
        mcutinv(i1,generator)=0d0
        if(i1.ge.1)tcutprocess(i1,generator)=0d0
      enddo
c angular cuts between beams and outgoing particles
      if(lsincludecuts.eq.3)then
        do i1=1,2
        do i2=3,nexternal(generator)
c ccutgen(i1,i2) = cos(thetacut) of angle between incoming particle i1
c               and outgoing particle i2
          if(ccutgen(i1,i2,generator).gt.0d0.and.
     *      ccutgen(i1,i2,generator).lt.1d0)then
c tcutprocess = tan(thetacut) of angle between incoming particle i1
c               and outgoing particle i2
            tcutprocess(2**(i1-1)+2**(i2-1),generator)=
     *        1d0/ccutgen(i1,i2,generator)**2-1d0
          endif
        enddo
        enddo
      endif
c invariant-mass cut
* determined from ecutgen,scutgen,ccutgen and external masses
c      write(*,*) 'initgenerator: mcmass = ',mcmass
c      write(*,*) 'initgenerator: idhep  = ',(idhep(i1,1),
c     &    i1=1,nexternal(generator))
      do i1=1,allbinary(0,generator)
        scutinv=0d0
        do i2=1,nexternal(generator)
          if(included(i1,2**(i2-1),nexternal(generator)))then
c CHANGED AD 30.03.06   (ok according to Markus)
c             scutinv=scutinv+mcmass(abs(idhep(i2,1)))
             scutinv=scutinv+sqrt(massext2(i2,generator))
           endif
        enddo
        scutinv=scutinv**2
        if(lsincludecuts.eq.3)then
          do i2=1,nexternal(generator)
          do i3=i2+1,nexternal(generator)
            if(included(i1,2**(i2-1),nexternal(generator)).and.
     *        included(i1,2**(i3-1),nexternal(generator)))then


              m2=mcmass(abs(idhep(i2,1)))
              m3=mcmass(abs(idhep(i3,1)))
              e2=max(ecutgen(i2,generator),m2)
              e3=max(ecutgen(i3,generator),m3)

c      write(*,*) 'initgenerator: included = ',i1,i2,i3,scutinv,
c     &            m2,m3,e2,e3,scutgen(i2,i3,generator),
c     &            ccutgen(i2,i3,generator)

              scutinv=scutinv-(m2+m3)**2
     *          +max(scutgen(i2,i3,generator),m2**2+m3**2+2d0*e2*e3
     *              -2d0*sqrt((e2**2-m2**2)*(e3**2-m3**2))
     *                *ccutgen(i2,i3,generator))

c      write(*,*) 'initgenerator: scutinv = ',scutinv

            endif
          enddo
          enddo
        endif
        mcutinv(i1,generator)=sqrt(scutinv)
        do i2=2,i1-1
          if(included(i1,i2,nexternal(generator)))then
            mcutinv(i1,generator)=max(mcutinv(i1,generator),
     *        mcutinv(i2,generator)+mcutinv(i1-i2,generator))
          endif
        enddo
      enddo

c      write(*,*) 'initgenerator: generator = ',generator
c      write(*,*) 'initgenerator: minvcut   = ',
c     &    lsincludecuts,scutgen(1,2,generator)
c      write(*,*) 'initgenerator: minvcut1   = ',
c     &    (mcutinv(i1,generator),i1=1,allbinary(0,generator))

c added AD 28.03.07
      if(lsincludecuts.ge.1)then
        mcutinv(allbinary(0,generator)-3,generator)=max(mtotcutgen,
     &      mcutinv(allbinary(0,generator)-3,generator))
      endif

c      write(*,*) 'initgenerator: minvcut2   = ',
c     &    (mcutinv(i1,generator),i1=1,allbinary(0,generator))

      sppmin=mcutinv(allbinary(0,generator)-3,generator)**2

c      write(*,*) 'initgenerator: sppmin   = ',sppmin

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                 c
c     initialize channels                                         c
c                                                                 c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
* add new channel
      channel=nchannel(generator)+1
c      write(*,*) 'initgenerator ch = ',channel,maxch
       if(channel.gt.maxch)then
         write(*,'(a28,i6)')' initgenerator: reset maxch >',
     *       channel
         stop
       endif
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                 c
c     loop over emitter and spectator                             c
c     assumes co=nexternal(generator)!!!!!!!!                     c
c                                                                 c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      do em=0,nexternal(generator)
      do sp=0,nexternal(generator)
      do co=0,nexternal(generator)
      n=0
c AD moved to init_subprocesses 04.12.06
c      lmap(em,sp,co,generator)=.false.
      if(em.eq.0.and.sp.eq.0.and.co.eq.0)then
* channel for original process
        n=nexternal(generator)
      elseif(lssub.ne.0.and.em.ne.sp.and.em.ne.co.and.sp.ne.co.and.
     *  em.gt.0.and.sp.gt.0.and.co.gt.0)then
* channels for subtraction terms
        if((pname(lhepnum(co)).eq.'ph '.or.
     &      pname(lhepnum(co)).eq.'gl ')
     &      .and.co.ge.3.and.
     *    vertex(lhepnum(em),-lhepnum(em),lhepnum(co),schannel).and.
     *    vertex(lhepnum(sp),-lhepnum(sp),lhepnum(co),schannel))then
          n=nexternal(generator)-1
          idhep(co,1)=0
        endif
      endif
      if(n.ne.0)then
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                 c
c     s-channel propagators                                       c
c                                                                 c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
*                                                                 *
*     if vertex  out1-out2-virt exists                            *
*     replace out1 by virt and discard out2 for next step         *
*                                                                 *
*******************************************************************
c number of s-channel propagators
      do maxns=0,n-3
        ns=0
        if(maxns.eq.0)goto 500
c ns'th s-channel propagator
 100    ns=ns+1
        out1(ns)=2
c decay product 1 for ns'th decay
 200    out1(ns)=out1(ns)+1
        out2(ns)=out1(ns)
c decay product 2 for ns'th decay
 300    out2(ns)=out2(ns)+1
        virt(ns)=-maxv-1
c virtual particle for ns'th decay
 400    virt(ns)=virt(ns)+1
* skip nonexisting virtual particles
        if(gname(virt(ns)).eq.'   '.and.virt(ns).lt.maxv)goto 400
* increase counters
        if(gname(virt(ns)).eq.'   '.and.virt(ns).eq.maxv)goto 1500
* skip antiparticles that equal particles
        if(virt(ns).lt.0.and.
     *    gname(virt(ns)).eq.gname(-virt(ns)))goto 400
c checking whether 3-particle vertex exists
        schannel=.true.
        if(.not.vertex(idhep(out1(ns),ns),idhep(out2(ns),ns),
     *    virt(ns),schannel))goto 1500
c checking last 3-particle vertex
c here in1(n-3) and in2(n-3) are equal to 1 and 2
c this statement could possibly be omitted since this vertex is checked
c later anyhow
         if(ns.eq.n-3.and..not.vertex(idhep(in1(ns),ns),
     *    idhep(in2(ns),ns),-virt(ns),schannel))goto 1500
c initializing next step
        do i2=1,n
          binary(i2,ns+1)=binary(i2,ns)
          idhep(i2,ns+1)=idhep(i2,ns)
        enddo
c combining particle out1 and out2 into new external particle out2
        binary(out1(ns),ns+1)=0
        idhep(out1(ns),ns+1)=0
        binary(out2(ns),ns+1)=
     *    binary(out1(ns),ns)+binary(out2(ns),ns)
        idhep(out2(ns),ns+1)=-virt(ns)
c find n s-channel propagator for ns < maxns
        if(ns.lt.maxns)goto 100     ! increase ns
 500    continue
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                 c
c     t-channel propagators                                       c
c                                                                 c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
*                                                                 *
*     if vertex  in1-out1-virt exists                             *
*     replace in1 by virt and discard out1 for next step          *
*                                                                 *
*******************************************************************
c number of t-channel propagators
        maxnt=n-3
c loop of t-channel propagators
        if(maxns.eq.maxnt)goto 1000
        nt=maxns
c nt'th t-channel propagator
 600    nt=nt+1
        in1(nt)=0
c incoming particle 1 for nt's process
 700    in1(nt)=in1(nt)+1
c incoming particle 2 for nt's process
        in2(nt)=3-in1(nt)
        out1(nt)=2
c outgoing particle 1 for nt's process
 800    out1(nt)=out1(nt)+1
c virtual particle for nt's process
        virt(nt)=-nmap                  ! (-nmap <-> -maxv-1)
 900    virt(nt)=virt(nt)+1
c replace maxv by nmap in following lines????
* skip nonexisting virtual particles
        if(gname(virt(nt)).eq.'   '.and.virt(nt).lt.maxv)goto 900
* increase counters
* check only for virt(nt).gt.maxv???
        if(gname(virt(nt)).eq.'   '.and.virt(nt).eq.maxv)goto 1300
* skip antiparticles that equal particles
        if(virt(nt).lt.0.and.
     *    gname(virt(nt)).eq.gname(-virt(nt)))goto 900
c avoid double counting of diagrams:
* first t-channel process with incoming particle 1
* next t-channel process with incoming particle 2
        if(nt.gt.maxns+1) then
          if((nt-1).eq.in1(nt))goto 1300
        endif
c checking whether 3-particle vertex in1-out1-virt exists
        schannel=.false.
        if(.not.vertex(idhep(in1(nt),nt),idhep(out1(nt),nt),
     *    virt(nt),schannel))goto 1300
c initializing next step
        do i2=1,n
          binary(i2,nt+1)=binary(i2,nt)
          idhep(i2,nt+1)=idhep(i2,nt)
        enddo
c combining particle in1 and out1 into new external particle in1
        binary(out1(nt),nt+1)=0
        idhep(out1(nt),nt+1)=0
        binary(in1(nt),nt+1)=binary(in1(nt),nt)+binary(out1(nt),nt)
        idhep(in1(nt),nt+1)=-virt(nt)
c find next t-channel propagator for nt < maxnt
        if(nt.lt.maxnt)goto 600    ! increase nt
 1000   continue
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                 c
c     checking last vertex                                        c
c                                                                 c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
* determine last existing outgoing line
        i1=n-2
        lastoutline = 0
        do i2=3,n
          if(idhep(i2,i1).ne.0) lastoutline=i2
        enddo
c        write(*,*) 'initgenerator: idhep(i,n-2) ',
c     &      (idhep(i2,n-2),i2=1,n)
        schannel=.true.
* check if last vertex exists
        if(gname(idhep(lastoutline,i1)).eq.'   '.or..not.
     *      vertex(idhep(1,i1),idhep(2,i1),
     &      idhep(lastoutline,i1),schannel))
     *      goto 1300
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                 c
c     initializing channels                                       c
c                                                                 c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c channel exists for process
        if (flag.eq.0) flag = -1
* number of invariants (without s?)
        ninv(channel,generator)=n-4
* number of t-channel processes
        nprocess(channel,generator)=maxnt-maxns
* number of decays
        ndecay(channel,generator)=maxns
* idhep of first s-channel propagator in channels without
* t-channel processes
        idhepfirst(channel,generator)=0
        if(nprocess(channel,generator).eq.0)
     *    idhepfirst(channel,generator)=abs(idhep(lastoutline,i1))
c initializing inv
c s-channel invariants corresponding to propagators
        do i1=1,maxns
* binary for i1'th s-channel invariant
          ininv(i1,channel,generator)=
     *      binary(out1(i1),i1)+binary(out2(i1),i1)
* idhep for i1'th s-channel invariant
          idhepinv(i1,channel,generator)=abs(virt(i1))
* power for i1'th s-channel invariant
          powerinv(i1,channel,generator)=power(abs(virt(i1)))
        enddo
c s-channel invariants corresponding to t-channel processes
        do i1=maxns+1,maxnt-1
          i2=maxns+maxnt-i1
* binary for i2'th out2 of t-channel process
          ininv(i2,channel,generator)=allbinary(co,generator)
     *      -binary(in1(i1),i1)-binary(in2(i1),i1)-binary(out1(i1),i1)
* idhep for i2'th
          idhepinv(i2,channel,generator)=0
* idhep for i2'th
          powerinv(i2,channel,generator)=0d0
        enddo
c minimal invariant-mass cuts for subroutine inv
        do i1=1,ninv(channel,generator)
        do i2=1,i1-1
          binary1=ininv(i1,channel,generator)
          binary2=ininv(i2,channel,generator)
          lmin(i2,i1,channel,generator)=.false.
* check whether inv i2 is relevant for (contained in) inv i1
          if(included(binary1,binary2,n))then
            lmin(i2,i1,channel,generator)=.true.
            do i3=1,i1-1
              binary3=ininv(i3,channel,generator)
              if(included(binary3,binary2,n)
     *          .and.i2.ne.i3)then
                lmin(i2,i1,channel,generator)=.false.
              endif
            enddo
          endif
        enddo
        enddo
c maximal invariant-mass cuts for subroutine inv
        do i1=1,ninv(channel,generator)
        do i2=1,i1-1
          binary1=allbinary(co,generator)-3-ininv(i1,channel,generator)
          binary2=ininv(i2,channel,generator)
          lmax(i2,i1,channel,generator)=.false.
          if(included(binary1,binary2,n))then
            lmax(i2,i1,channel,generator)=.true.
            do i3=1,i1-1
              binary3=ininv(i3,channel,generator)
              if(included(binary3,binary2,n)
     *          .and.i2.ne.i3)then
                lmax(i2,i1,channel,generator)=.false.
              endif
            enddo
          endif
        enddo
        enddo
c initializing process
        do i1=maxns+1,maxnt
          i2=i1-maxns
          in1process(i2,channel,generator)=allbinary(co,generator)
     *      -binary(in1(i1),i1)
          in2process(i2,channel,generator)=allbinary(co,generator)
     *      -binary(in2(i1),i1)
          out1process(i2,channel,generator)=binary(out1(i1),i1)
          out2process(i2,channel,generator)=allbinary(co,generator)
     *      -binary(in1(i1),i1)-binary(in2(i1),i1)-binary(out1(i1),i1)
          inprocess(i2,channel,generator)=allbinary(co,generator)
     *      -binary(in1(i1),i1)-binary(in2(i1),i1)
          virtprocess(i2,channel,generator)=allbinary(co,generator)
     *      -binary(in1(i1),i1)-binary(out1(i1),i1)
          idhepprocess(i2,channel,generator)=abs(virt(i1))
          powerprocess(i2,channel,generator)=power(abs(virt(i1)))
        enddo
c initializing decay
        do i1=1,maxns
* reverse order to get invariants with few momenta first
          i2=maxns-i1+1
          indecay(i2,channel,generator)=
     *      binary(out1(i1),i1)+binary(out2(i1),i1)
          out1decay(i2,channel,generator)=binary(out1(i1),i1)
          out2decay(i2,channel,generator)=binary(out2(i1),i1)
        enddo
c initializing map

        lmap(em,sp,co,generator)=.true.
        emmap(channel,generator)=em
        spmap(channel,generator)=sp
        comap(channel,generator)=co
c        if(em.eq.0.and.sp.eq.0.and.co.eq.0)then
c          write(*,*) 'initgen lmap ',
c     &        em,sp,co,generator,lmap(em,sp,co,generator)
c        endif

c test output of found channel
c     end test


c checking whether generator already exists
        do i1=1,channel-1
          if(comparechannel(i1,channel,smap,generator))then
c ask Markus why necessary
c            write(*,*) 'initgenerator: tcutprocess',
c     &          allbinary(co,generator)
c     &          , virtprocess(1,channel,generator)
c     &          ,channel
c ensures that if a t-channel cut matters this version of the channel is used
c comparechannel does not compare these cuts
            if(nprocess(channel,generator).gt.0) then
              if (tcutprocess(allbinary(co,generator)
     &         -virtprocess(1,channel,generator),generator).ne.0d0.or.
     &         tcutprocess(virtprocess(1,channel,generator),generator)
     &         .ne.0d0) then
                call copychannel(channel,i1,generator)
              endif
            endif
            goto 1300
          endif
        enddo
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                 c
c     output new channel                                          c
c                                                                 c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c        nout = 6
        if(nout.ne.0.and.lnoutgen.ge.2)then
          write(nout,*)
          write(nout,'(a12,i6)')' channel   =',channel
          if(em.ne.0.or.sp.ne.0.or.co.ne.0)then
            write(nout,'(a12,i6)')' emitter   =',em
            write(nout,'(a12,i6)')' spectator =',sp
            write(nout,'(a12,i6)')' photon    =',co
          endif
          do i1=1,maxns
            write(nout,'(a32,i6," +",i6," ->",i6)')
     *        '   s channel: ('//pname(-idhep(out1(i1),i1))//','//
     *        pname(-idhep(out2(i1),i1))//','//
     *        pname(-idhep(out2(i1),i1+1))//'), ',
     *        pid(binary(out1(i1),i1)),pid(binary(out2(i1),i1)),
     *        pid(binary(out2(i1),i1+1))
          enddo
          do i1=maxns+1,maxnt
            write(nout,'(a32,i6," +",i6," ->",i6)')
     *        '   t channel: ('//pname(idhep(in1(i1),i1))//','//
     *        pname(-idhep(out1(i1),i1))//','//
     *        pname(idhep(in1(i1),i1+1))//'), ',
     *        pid(binary(in1(i1),i1)),pid(binary(out1(i1),i1)),
     *        pid(binary(in1(i1),i1+1))
          enddo
            write(nout,'(a32,i6," +",i6," ->",i6)')
     *        '   last vertex: ('//pname(idhep(1,n-2))//','//
     *        pname(idhep(2,n-2))//','//
     *        pname(-idhep(lastoutline,n-2))//'), ',
     *        pid(binary(1,n-2)),pid(binary(2,n-2)),
     *        pid(binary(lastoutline,n-2))
        endif
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                 c
c     end of loops                                                c
c                                                                 c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
        flag = 1
        channel=channel+1
c        write(*,*) 'initgenerator ch3 = ',channel,maxch
        if(channel.gt.maxch)then
          write(*,'(a28,i6)')' initgenerator: reset maxch >',
     *        channel
          stop
        endif
 1300   if(maxns.lt.maxnt)then
* lt -> le???
 1400     if(virt(nt).lt.nmap-1)goto 900  ! increase virt(nt)
                                          ! (nmap-1 <-> maxv)
          if(out1(nt).lt.n)goto 800       ! increase out1(nt)
          if(in1(nt).lt.2)goto 700        ! increase in1(nt)
          if(nt.gt.maxns+1)then           ! go one nt back
            nt=nt-1
            goto 1400
          endif
        endif
c end of t-channel propagators
 1500   if(maxns.gt.0)then
 1600     if(virt(ns).lt.maxv)goto 400    ! increase virt(ns)
          if(out2(ns).lt.n)goto 300       ! increase out2(ns)
          if(out1(ns).lt.n-1)goto 200     ! increase out1(ns)
          if(ns.gt.1)then                 ! go one ns back
            ns=ns-1
            goto 1600
          endif
        endif
c end of s-channel propagators
      enddo
c number of channels
      if(channel.eq.nchannel(generator)+1) then
c        if (flag.eq.0) then
c no channel found for process
c          write(*,'(a47,i2)')
c     *        ' initgenerator: no channel found for generator '
c     &        ,generator
c        else if (flag.eq.-1) then
c          write(*,'(a50,i2)')
c     *        ' initgenerator: no new channel found for generator '
c     &        ,generator
c        endif
      else
        flag = 1
        nchannel(generator)=channel-1
c        if(nchannel(generator).gt.maxch)then
c          write(*,'(a28,i6)')' initgenerator: reset maxch >',
c     *        nchannel(generator)
c          stop
c        endif
      endif
c      write(nout,'(a12,i6)')' initgenerator:nchannel   =',
c     &    nchannel(generator)
c      if(nchannel(generator).eq.0)then
c        write(*,'(a47,i2)')
c     *    ' initgenerator: no channel found for generator ',generator
c        stop
c      endif

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                 c
c     initialize densities and reduce redundancies                c
c                                                                 c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c initializing inv
      maxinv(generator)=0
      do channel=1,nchannel(generator)
      do ns=1,ninv(channel,generator)
        numinv(ns,channel,generator)=0
        do i1=1,channel-1
        do i2=1,ninv(i1,generator)
          if(compareinv(ns,channel,i2,i1,generator))
     *      numinv(ns,channel,generator)=numinv(i2,i1,generator)
        enddo
        enddo
        if(numinv(ns,channel,generator).eq.0)then
          if(maxinv(generator).lt.maxiv) then
            maxinv(generator)=maxinv(generator)+1
          else
            write(*,'(a29,i6)')' initgenerator: reset maxiv >',
     *          maxinv(generator)
          stop
          endif
          numinv(ns,channel,generator)=maxinv(generator)
          nsinv(maxinv(generator),generator)=ns
          chinv(maxinv(generator),generator)=channel
        endif
      enddo
      enddo
c process
      maxprocess(generator)=0
      do channel=1,nchannel(generator)
      do nt=1,nprocess(channel,generator)
        numprocess(nt,channel,generator)=0
        do i1=1,channel-1
        do i2=1,nprocess(i1,generator)
          if(compareprocess(nt,channel,i2,i1,generator))
     *      numprocess(nt,channel,generator)=numprocess(i2,i1,generator)
        enddo
        enddo
        if(numprocess(nt,channel,generator).eq.0)then
          if(maxprocess(generator).lt.maxpr) then
            maxprocess(generator)=maxprocess(generator)+1
          else
            write(*,'(a29,i6)')' initgenerator: reset maxpr >',
     *          maxprocess(generator)
            stop
          endif
          numprocess(nt,channel,generator)=maxprocess(generator)
          ntprocess(maxprocess(generator),generator)=nt
          chprocess(maxprocess(generator),generator)=channel
        endif
      enddo
      enddo
c decay
      maxdecay(generator)=0
      do channel=1,nchannel(generator)
      do ns=1,ndecay(channel,generator)
        numdecay(ns,channel,generator)=0
        do i1=1,channel-1
        do i2=1,ndecay(i1,generator)
          if(comparedecay(ns,channel,i2,i1,generator))
     *      numdecay(ns,channel,generator)=numdecay(i2,i1,generator)
        enddo
        enddo
        if(numdecay(ns,channel,generator).eq.0)then
          if(maxdecay(generator).lt.maxdc) then
            maxdecay(generator)=maxdecay(generator)+1
          else
            write(*,'(a29,i6)')' initgenerator: reset maxdc >',
     *          maxdecay(generator)
            stop
          endif
          numdecay(ns,channel,generator)=maxdecay(generator)
          nsdecay(maxdecay(generator),generator)=ns
          chdecay(maxdecay(generator),generator)=channel
        endif
      enddo
      enddo
      endif
      enddo
      enddo
      enddo
      if(nout.ne.0.and.lnoutgen.ge.10)then
c output
        write(nout,'(a)')' '
        write(nout,'(a23,i2)')' Phase-space generator ',generator
        write(nout,'(" Number of channels            =",i6)')
     *    nchannel(generator)
        write(nout,'(" Calculation of invariants     =",i6)')
     *    maxinv(generator)
        write(nout,'(" Calculation of 2->2 processes =",i6)')
     *    maxprocess(generator)
        write(nout,'(" Calculation of 1->2 decays    =",i6)')
     *    maxdecay(generator)
      endif
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                 c
c     initialize adaptive optimization                            c
c                                                                 c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      do i1=0,maxo
        nopt(i1,generator)=0
      enddo
      nopt(1,generator)=1000
      nopt(2,generator)=2000
      nopt(3,generator)=4000
      nopt(4,generator)=8000
      nopt(5,generator)=16000
      nopt(6,generator)=32000
      nopt(7,generator)=64000
      nopt(8,generator)=128000
      do i1=0,maxo
        nopt(i1,generator)=nopt(i1,generator)*nchannel(generator)
c added 18.06.07
     &      *freqgencon(generator,1)
      enddo
      opt(generator)=1
      betaopt(0,generator)=0d0
      do i1=1,nchannel(generator)
        wi(i1,generator)=0d0
        alphaopt(i1,1,generator)=1d0/nchannel(generator)
        betaopt(i1,generator)=dble(i1)/nchannel(generator)
      enddo
      alphamin=0.01d0
c      alphamin=0d0

c      write(*,*) 'initgenerator: generat =',generator
c      write(*,*) 'initgenerator: channel =',nchannel(generator)
c      write(*,*) 'initgenerator: betaopt =',betaopt

c      write(*,*) 'initgenerator: out ',generator,
c     &    lmap(0,0,0,generator)

      end

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                 c
c     check whether generator already exists                      c
c                                                                 c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      function comparechannel(ch1,ch2,lsmap,generator)
      implicit none

#include "mcparams.h"
#include "mccommon.h"

c local variables
      integer i1,i2,ch1,ch2,generator,prop1,prop2,lsmap,idhep1,idhep2
      integer binary1,binary2,co
c      logical comparechannel,same,exist
      logical comparechannel,exist

      comparechannel=.false.
c checking emitter, spectator, and photon
      if(emmap(ch1,generator).ne.emmap(ch2,generator))return
      if(spmap(ch1,generator).ne.spmap(ch2,generator))return
      if(comap(ch1,generator).ne.comap(ch2,generator))return
      co=comap(ch1,generator)
c checking number of s-channel propagators
      prop1=ndecay(ch1,generator)
      do i1=1,ndecay(ch1,generator)
        idhep1=idhepinv(i1,ch1,generator)
        if(mcmass(idhep1).eq.0d0.and.mcwidth(idhep1).eq.0d0.and.
     *    power(idhep1).eq.0d0)prop1=prop1-1
      enddo
      prop2=ndecay(ch2,generator)
      do i1=1,ndecay(ch2,generator)
        idhep2=idhepinv(i1,ch2,generator)
        if(mcmass(idhep2).eq.0d0.and.mcwidth(idhep2).eq.0d0.and.
     *    power(idhep2).eq.0d0)prop2=prop2-1
      enddo
c checking s-channel propagators
      if(prop1.eq.prop2)then
        do i1=1,ndecay(ch1,generator)
          binary1=ininv(i1,ch1,generator)
          idhep1=idhepinv(i1,ch1,generator)
          if(mcmass(idhep1).ne.0d0.or.mcwidth(idhep1).ne.0d0.or.
     *      power(idhep1).ne.0d0)then
            exist=.false.
            do i2=1,ndecay(ch2,generator)
              binary2=ininv(i2,ch2,generator)
              idhep2=idhepinv(i2,ch2,generator)
              if(binary1.eq.binary2.and.
     *          mcmass(idhep1).eq.mcmass(idhep2).and.
     *          mcwidth(idhep1).eq.mcwidth(idhep2).and.
     *          power(idhep1).eq.power(idhep2))exist=.true.
            enddo
            if(.not.exist)return
          endif
        enddo
      else
        return
      endif
c checking number of t-channel propagators
      prop1=nprocess(ch1,generator)
      do i1=1,nprocess(ch1,generator)
        idhep1=idhepprocess(i1,ch1,generator)
        if(mcmass(idhep1).eq.0d0.and.mcwidth(idhep1).eq.0d0.and.
     *    power(idhep1).eq.0d0)prop1=prop1-1
      enddo
      prop2=nprocess(ch2,generator)
      do i1=1,nprocess(ch2,generator)
        idhep2=idhepprocess(i1,ch2,generator)
        if(mcmass(idhep2).eq.0d0.and.mcwidth(idhep2).eq.0d0.and.
     *    power(idhep2).eq.0d0)prop2=prop2-1
      enddo
c checking t-channel propagators
      if(prop1.eq.prop2)then
        do i1=1,nprocess(ch1,generator)
          binary1=min(virtprocess(i1,ch1,generator),
     *      allbinary(co,generator)-virtprocess(i1,ch1,generator))
          idhep1=idhepprocess(i1,ch1,generator)
          if(mcmass(idhep1).ne.0d0.or.mcwidth(idhep1).ne.0d0.or.
     *      power(idhep1).ne.0d0)then
            exist=.false.
            do i2=1,nprocess(ch2,generator)
              binary2=min(virtprocess(i2,ch2,generator),
     *          allbinary(co,generator)-virtprocess(i2,ch2,generator))
              idhep2=idhepprocess(i2,ch2,generator)
              if(binary1.eq.binary2.and.
     *          mcmass(idhep1).eq.mcmass(idhep2).and.
     *          mcwidth(idhep1).eq.mcwidth(idhep2).and.
     *          power(idhep1).eq.power(idhep2))then
                  exist=.true.
              endif
            enddo
            if(.not.exist)return
          endif
        enddo
      else
        return
      endif
c checking first s-channel propagator
      if(lsmap.eq.0.and.nprocess(ch1,generator).eq.0.and.
     *  nprocess(ch2,generator).eq.0)then
        idhep1=idhepfirst(ch1,generator)
        idhep2=idhepfirst(ch2,generator)
        if(mcmass(idhep1).ne.mcmass(idhep2).or.
     *    mcwidth(idhep1).ne.mcwidth(idhep2).or.
     *    power(idhep1).ne.power(idhep2))return
      endif
      comparechannel=.true.
      end

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                 c
c     copy channel ch1 to channel ch2                             c
c                                                                 c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine copychannel(ch1,ch2,generator)
      implicit none

#include "mcparams.h"
#include "mccommon.h"

c local variables
      integer i1,i2,ch1,ch2,generator

      do i1=1,maxe
c inv
        powerinv(i1,ch2,generator)=powerinv(i1,ch1,generator)
        ininv(i1,ch2,generator)=ininv(i1,ch1,generator)
        idhepinv(i1,ch2,generator)=idhepinv(i1,ch1,generator)
        idhepfirst(ch2,generator)=idhepfirst(ch1,generator)
        ninv(ch2,generator)=ninv(ch1,generator)
        do i2=1,maxe
          lmin(i1,i2,ch2,generator)=lmin(i1,i2,ch1,generator)
          lmax(i1,i2,ch2,generator)=lmax(i1,i2,ch1,generator)
        enddo
c decay
        indecay(i1,ch2,generator)=indecay(i1,ch1,generator)
        out1decay(i1,ch2,generator)=out1decay(i1,ch1,generator)
        out2decay(i1,ch2,generator)=out2decay(i1,ch1,generator)
        ndecay(ch2,generator)=ndecay(ch1,generator)
c process
        powerprocess(i1,ch2,generator)=powerprocess(i1,ch1,generator)
        in1process(i1,ch2,generator)=in1process(i1,ch1,generator)
        in2process(i1,ch2,generator)=in2process(i1,ch1,generator)
        out1process(i1,ch2,generator)=out1process(i1,ch1,generator)
        out2process(i1,ch2,generator)=out2process(i1,ch1,generator)
        inprocess(i1,ch2,generator)=inprocess(i1,ch1,generator)
        virtprocess(i1,ch2,generator)=virtprocess(i1,ch1,generator)
        idhepprocess(i1,ch2,generator)=idhepprocess(i1,ch1,generator)
        nprocess(ch2,generator)=nprocess(ch1,generator)
      enddo
c mapping
      emmap(ch2,generator)=emmap(ch1,generator)
      spmap(ch2,generator)=spmap(ch1,generator)
      comap(ch2,generator)=comap(ch1,generator)
      end

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                 c
c     check whether two 1->2 decays are equal                     c
c                                                                 c
c     written by Markus Roth                                      c
c                                                                 c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      function comparedecay(ns1,ch1,ns2,ch2,generator)
      implicit none

#include "mcparams.h"
#include "mccommon.h"

c local variables
      integer ns1,ns2,ch1,ch2,generator
      logical comparedecay

      comparedecay=.false.
      if(emmap(ch1,generator).ne.emmap(ch2,generator))return
      if(spmap(ch1,generator).ne.spmap(ch2,generator))return
      if(comap(ch1,generator).ne.comap(ch2,generator))return
      if(indecay(ns1,ch1,generator).ne.indecay(ns2,ch2,generator))
     *  return
      if(out1decay(ns1,ch1,generator).ne.out1decay(ns2,ch2,generator)
     *  .and.out2decay(ns1,ch1,generator).ne.
     *  out2decay(ns2,ch2,generator))return
      comparedecay=.true.
      end

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                 c
c     check whether two 2->2 processes are equal                  c
c                                                                 c
c     written by Markus Roth                                      c
c                                                                 c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      function compareprocess(ns1,ch1,ns2,ch2,generator)
      implicit none
c local variables

#include "mcparams.h"
#include "mccommon.h"

      integer ns1,ns2,ch1,ch2,generator,idhep1,idhep2,co
      logical compareprocess

      compareprocess=.false.
      if(emmap(ch1,generator).ne.emmap(ch2,generator))return
      if(spmap(ch1,generator).ne.spmap(ch2,generator))return
      if(comap(ch1,generator).ne.comap(ch2,generator))return
      co=comap(ch1,generator)
      if(inprocess(ns1,ch1,generator).ne.inprocess(ns2,ch2,generator))
     *  return
      if(virtprocess(ns1,ch1,generator).ne.
     *  virtprocess(ns2,ch2,generator).and.
     *  virtprocess(ns1,ch1,generator).ne.
     *  allbinary(co,generator)-virtprocess(ns2,ch2,generator))return
      if(powerprocess(ns1,ch1,generator).ne.
     *  powerprocess(ns2,ch2,generator))return
      idhep1=idhepprocess(ns1,ch1,generator)
      idhep2=idhepprocess(ns2,ch2,generator)
      if(mcmass(idhep1).ne.mcmass(idhep2).or.
     *    mcwidth(idhep1).ne.mcwidth(idhep2))
     *  return
      compareprocess=.true.
      end

***********************************************************************
      function compareinv(ns1,ch1,ns2,ch2,generator)
***********************************************************************
*     checks whether calculation of invariants ns1 and ns2            *
*     in channels ch1 and ch2 are equal                               *
*---------------------------------------------------------------------*
*     by Markus Roth         last changed   00.00.06   Ansgar Denner  *
***********************************************************************
      implicit none

#include "mcparams.h"
#include "mccommon.h"

      integer i1,i2,ns1,ns2,ch1,ch2,generator,idhep1,idhep2
      logical compareinv,included

      compareinv=.false.
      if(emmap(ch1,generator).ne.emmap(ch2,generator))return
      if(spmap(ch1,generator).ne.spmap(ch2,generator))return
      if(comap(ch1,generator).ne.comap(ch2,generator))return
      if(ininv(ns1,ch1,generator).ne.ininv(ns2,ch2,generator))return
      if(powerinv(ns1,ch1,generator).ne.powerinv(ns2,ch2,generator))
     *  return
      idhep1=idhepinv(ns1,ch1,generator)
      idhep2=idhepinv(ns2,ch2,generator)
      if(mcmass(idhep1).ne.mcmass(idhep2).or.
     *    mcwidth(idhep1).ne.mcwidth(idhep2))
     *  return
      do i1=1,ns1-1
        if(lmin(i1,ns1,ch1,generator))then
          included=.false.
          do i2=1,ns2-1
            if(lmin(i2,ns2,ch2,generator).and.
     *        ininv(i1,ch1,generator).eq.ininv(i2,ch2,generator))
     *        included=.true.
          enddo
          if(.not.included)return
        endif
        if(lmax(i1,ns1,ch1,generator))then
          included=.false.
          do i2=1,ns2-1
            if(lmax(i2,ns2,ch2,generator).and.
     *        ininv(i1,ch1,generator).eq.ininv(i2,ch2,generator))
     *        included=.true.
          enddo
          if(.not.included)return
        endif
      enddo
      do i2=1,ns2-1
        if(lmin(i2,ns2,ch2,generator))then
          included=.false.
          do i1=1,ns1-1
            if(lmin(i1,ns1,ch1,generator).and.
     *        ininv(i1,ch1,generator).eq.ininv(i2,ch2,generator))
     *        included=.true.
          enddo
          if(.not.included)return
        endif
        if(lmax(i2,ns2,ch2,generator))then
          included=.false.
          do i1=1,ns1-1
            if(lmax(i1,ns1,ch1,generator).and.
     *        ininv(i1,ch1,generator).eq.ininv(i2,ch2,generator))
     *        included=.true.
          enddo
          if(.not.included)return
        endif
      enddo
      compareinv=.true.
      end

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                 c
c     check if binary2 is included in binary1                     c
c                                                                 c
c     written by Markus Roth                                      c
c                                                                 c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
***********************************************************************
      function included(binary1,binary2,nexternal)
***********************************************************************
*     checks whether  binary2 is included in binary1                  *
*     in the sense that for the coefficients of each power of 2:      *
*     0 and 1 are included in 1                                       *
*     0 is included in 0                                              *
*     but 1 is not included in 0                                      *
*---------------------------------------------------------------------*
*     by Markus Roth         last changed   00.00.06   Ansgar Denner  *
***********************************************************************
      implicit none
c local variables
      integer binary1,binary2,i1,b1,b2,nexternal
      logical included
      included=.false.
      do i1=1,nexternal
        b1=binary1/2**(i1-1)
        b2=binary2/2**(i1-1)
        if(2*(b2/2).ne.b2.and.2*(b1/2).eq.b1)return
      enddo
      included=.true.
      end

***********************************************************************
      subroutine inv(random,s,g,mass,width,lpower,smin,smax,step,switch)
***********************************************************************
*     mapping of s-channel propagators                                *
*     input:                                                          *
*       random: random number                                         *
*       mass,width: of corresponding resonance                        *
*       lpower: controls mapping                                      *
*       smin,smax: lower and upper limit for s                        *
*     output:                                                         *
*       s: invariant  for step=1 and switch = 0                       *
*       g: density    for step=2 and switch = 0                       *
*       switch > 0 for error                                          *
*---------------------------------------------------------------------*
*     by Markus Roth         last changed   06.11.12   Ansgar Denner  *
***********************************************************************
      implicit none
c local variable
      real*8 random,s,g,mass,lmass2,width,width2,lpower
      real*8 smax,smin,omax,omin,mapinv,jacobianinv,denum
      real*8 mmid,domid,domax,widthfac
      integer step,switch

#include "mcparams.h"
#include "mccommon.h"

      data mmid /160d0/,widthfac /4.922d-7/
c     mmid approx 2*mw
c     widthfac approx 3*GF/(16 sqrt(2) pi)
c     gamma approx widthfac*mh^3 for large mh

c      write(*,*) 'inv: input',mass,width,smin,smax,lpower

      if(switch.gt.0)return
      lmass2=mass*mass
      if(smax.le.smin)then
        switch=3
        if(numout.lt.maxout.and.lnoutmc.ge.5)then
 100        format(/" event = ",i9,", gen = ",i4,", channel = ",i4,
     &        ", exit = ",i4,(a))
            write(nout,100) nevt,cogen,cocha,switch,
     &        ', inv: smin >= smax'
          write(nout,'(6x," inv: smin >= smax",2i3)')cogen,cocha
          write(nout,'(6x,"smin=",d16.10," smax=",d16.10)')smin,smax
          numout=numout+1
        endif
        return
      endif
      if(smin.lt.0d0)then
        switch=4
        if(numout.lt.maxout.and.lnoutmc.ge.5)then
            write(nout,100) nevt,cogen,cocha,switch,
c          write(nout,'(" event = ",i9,",",(a))') nevt,
     &          ' inv: smin < 0'
          numout=numout+1
        endif
        return
      endif
      if(width.gt.0d0)then
        mmid=sqrt(lmass2+sqrt(mass*width*(1d0/widthfac-mass*width)))
        mmid=max(160d0,mmid)
        if(mass.ne.mcm(25).or.shbw.ne.1.or.smax.le.mmid*mmid)then
          width2=width*width
          omax=datan((smax-lmass2)/mass/width)
          omin=datan((smin-lmass2)/mass/width)
          if(step.eq.1)then

c          write(*,*) 'inv: use random=',random

            s=lmass2+mass*width*dtan(random*(omax-omin)+omin)
          elseif(step.eq.2)then
            denum=(omax-omin)*((s-lmass2)**2+lmass2*width2)
            if(denum.gt.0d0)then
              g=g*mass*width/denum
            else
              switch=5
              if(numout.lt.maxout.and.lnoutmc.ge.5)then
                write(nout,100) nevt,cogen,cocha,switch,
     &              ' inv: 1/density <= 0 (width>0)'
                write(nout,'(6x,"  inv: 1/density <= 0 (width>0)",2i3)')
     &              cogen,cocha
                write(nout,'(6x,"smin=",d16.10," smax=",d16.10)')smin,
     &              smax
                write(nout,'(6x,"omin=",d16.10," omax=",d16.10,d18.10)')
     &              omin,omax,omax-omin
                numout=numout+1
              endif
              return
            endif
          endif
        else  !  special Higgs propagator with s-dependent width in numerator
          width2=width*width
          if(mmid*mmid.gt.smin) then
            omin=datan((smin-lmass2)/(mass*width))
            domid=datan((mmid*mmid-lmass2)/(mass*width))-omin
            domax=domid+2d0*widthfac*mass*(sqrt(smax)-mmid)

            if(step.eq.1)then
              if(random*domax.gt.domid)then
                s=(random*domax-domid)/(2d0*widthfac*mass)+mmid
                s=s*s
              else
                s=lmass2+mass*width*dtan(random*domax+omin)
              endif
            elseif(step.eq.2)then
              if(s.gt.mmid*mmid)then
                denum=sqrt(s)*domax/(widthfac*mass)
              else
                denum=domax*((s-lmass2)**2+lmass2*width2)/(mass*width)
              endif
              if(denum.gt.0d0)then
                g=g/denum
              else
                switch=5
                if(numout.lt.maxout.and.lnoutmc.ge.5)then
                  write(nout,100) nevt,cogen,cocha,switch,
     &                ' inv: 1/density <= 0 (width>0)'
                  write(nout,'(6x,"  inv: 1/density <= 0 (width>0)",
     &                   2i3)')
     &                cogen,cocha
                  write(nout,'(6x,"smin=",d16.10," smax=",d16.10)')
     &                smin,smax
                  write(nout,'(6x,"omin=",d16.10," omax=",d16.10,
     &                   d18.10)')
     &                omin,omax,omax-omin
                  numout=numout+1
                endif
                return
              endif
            endif
          else
            omin=2d0*widthfac*mass*sqrt(smin)
            domax=2d0*widthfac*mass*sqrt(smax)-omin

            if(step.eq.1)then
              s=(random*domax)/(2d0*widthfac*mass)+sqrt(smin)
              s=s*s
            elseif(step.eq.2)then
              denum=sqrt(s)*domax/(widthfac*mass)
              if(denum.gt.0d0)then
                g=g/denum
              else
                switch=5
                if(numout.lt.maxout.and.lnoutmc.ge.5)then
                  write(nout,100) nevt,cogen,cocha,switch,
     &                ' inv: 1/density <= 0 (width>0)'
                  write(nout,'(6x,"  inv: 1/density <= 0 (width>0)",
     &                   2i3)')
     &                cogen,cocha
                  write(nout,'(6x,"smin=",d16.10," smax=",d16.10)')
     &                smin,smax
                  write(nout,'(6x,"omin=",d16.10," omax=",d16.10,
     &                   d18.10)')
     &                omin,omax,omax-omin
                  numout=numout+1
                endif
                return
              endif
            endif
          endif
        endif
      else
        if(step.eq.1)then

c          write(*,*) 'inv: use random=',random

          s=mapinv(random,lpower,lmass2,smin,smax,switch)
        elseif(step.eq.2)then
          denum=jacobianinv(lpower,lmass2,s,smin,smax,switch)
          if(denum.gt.0d0)then
            g=g/denum
          else
            switch=6
            if(numout.lt.maxout.and.lnoutmc.ge.5)then
              write(nout,100) nevt,cogen,cocha,switch,
c                write(nout,'(" event = ",i9,",",(a))') nevt,
     &            ' inv: 1/density <= 0 (width=0)'
              write(nout,'(6x,"  inv: 1/density <= 0 (width>0)",2i3)')
     &            cogen,cocha
              write(nout,'(6x,"smin=",d16.10," smax=",d16.10)')smin,smax
             endif
            numout=numout+1
            return
          endif
        endif
      endif
c      write(*,*) 'inv ginv',g,denum,lpower,lmass2,s,smin,smax
      end

***********************************************************************
      subroutine process(random,q1,q2,p1,p2,qt,g,mass,width,lpower,
     *  s,s1,s2,t,t1,t2,tcut1,tcut2,step,switch)
***********************************************************************
*     mapping of  2->2 particle process with t-channel propagator     *
*     input:                                                          *
*       random(2): random numbers for angles                          *
*       q1,q2: incoming momenta                                       *
*       s: invariant mass of initial state                            *
*       s1,s2: invariant masses of incoming momenta                   *
*       t1,t2: invariant masses of outgoing momenta                   *
*       mass,width of corresponding virtual particle                  *
*       lpower: controls mapping of propagator                        *
*       tcut1,tcut2: cut parameters for tan^2(theta1), tan^2(theta2)  *
*     output:                                                         *
*       p1,p2: outgoing momenta           for step=1 and switch = 0   *
*       qt: momentum of propagator q1-p1  for step=1 and switch = 0   *
*       t: invariant of propagator        for step=1 and switch = 0   *
*       g: density                        for step=2 and switch = 0   *
*       switch > 0 for error                                          *
*     remarks                                                         *
*       q1+q2 must be in z-direction                                  *
*             with allowed uncertainty determined by techparam(3)     *
*       resulting t and tmax must be necotive                         *
*             with allowed uncertainty determined by techparam(2)     *
*---------------------------------------------------------------------*
*     by Markus Roth         last changed   14.11.06   Ansgar Denner  *
***********************************************************************
      implicit none

#include "mcparams.h"
#include "mccommon.h"

c local variable
c      real*8 pi,mass2
      real*8 random(2),q1(0:3),q2(0:3),p1(0:3),p2(0:3),qt(0:3)
      real*8 g,mass,width,lpower,s1,s2,t1,t2,tmin,tmax,tcut1,tcut2
      real*8 lambdas,lambdat,lmass2,width2,t,phi,cost,s,roots
      real*8 gamma2,beta,gs,d,lambda,be,omeco,omin,omax
      real*8 k1(0:3),q(0:3),mapinv,jacobianinv,denum,cmax,cmax1,cmax2
      integer i1,step,switch
c mcoutput
c      integer nout,numout,maxout,noutgen,nsteps
c mctechparam
c      real*8 techparam(4)
c      common/mcoutput/nout,numout,maxout,noutgen,nsteps
c      common/mctechparam/techparam

      if(switch.gt.0)return
c      pi=4d0*datan(1d0)
      if(tcut1.ne.0d0.or.tcut2.ne.0d0)then
        beta=(q1(3)+q2(3))/(q1(0)+q2(0))
        gamma2=1d0/(1d0-beta*beta)
        lambda=(s-s1-s2)**2-4d0*s1*s2
        if(dabs(q1(1)+q2(1)).gt.techparam(3).and.
     *    dabs(q1(2)+q2(2)).gt.techparam(3))then
          switch=7
          if(numout.lt.maxout.and.lnoutmc.ge.5)then
 100        format(/" event = ",i9,", gen = ",i4,", channel = ",i4,
     &          ", exit = ",i4,(a))
            write(nout,100) nevt,cogen,cocha,switch,
     &          ', process: dabs(q1(1:2)+q2(1:2)) > acc '
            numout=numout+1
          endif
          return
        elseif(lambda.lt.0d0)then
          switch=8
          if(numout.lt.maxout.and.lnoutmc.ge.5)then
            write(nout,100) nevt,cogen,cocha,switch,
c                write(nout,'(" event = ",i9,",",(a))') nevt,
     &          ' process: lambda < 0'
            numout=numout+1
          endif
          return
        else
          if(tcut1.ne.0d0)then    ! tcut1=tan**2(theta_1)
            be=beta*dsign(1d0,q1(3))
            gs=be*(s+s1-s2)/sqrt(lambda)
            d=1d0+gamma2*(1d0-gs*gs)*tcut1
            if(d.lt.0d0)then
              switch=9
              if(numout.lt.maxout.and.lnoutmc.ge.5)then
                write(nout,100) nevt,cogen,cocha,switch,
c                write(nout,'(" event = ",i9,",",(a))') nevt,
     &              ' process: d < 0'
                numout=numout+1
              endif
              return
            else
              cmax1=(-gs*gamma2*tcut1+sqrt(d))/(1d0+gamma2*tcut1)
            endif
          else
            cmax1=1d0
          endif
          if(tcut2.ne.0d0)then    ! tcut2=tan**2(theta_2)
            be=beta*dsign(1d0,q2(3))
            gs=be*(s+s2-s1)/sqrt(lambda)
            d=1d0+gamma2*(1d0-gs*gs)*tcut2
            if(d.lt.0d0)then
              switch=9
              if(numout.lt.maxout.and.lnoutmc.ge.5)then
                write(nout,100) nevt,cogen,cocha,switch,
c                write(nout,'(" event = ",i9,",",(a))') nevt,
     &              ' process: d < 0'
                numout=numout+1
              endif
              return
            else
              cmax2=(-gs*gamma2*tcut2+sqrt(d))/(1d0+gamma2*tcut2)
            endif
          else
            cmax2=1d0
          endif
        endif
        cmax=min(cmax1,cmax2)
      else
        cmax=1d0
      endif
      lmass2=mass*mass
      lambdas=(s-s1-s2)*(s-s1-s2)-4d0*s1*s2
      if(lambdas.gt.0d0)then
        lambdas=sqrt(lambdas)
      else
        switch=10
        if(numout.lt.maxout.and.lnoutmc.ge.5)then
          write(nout,100) nevt,cogen,cocha,switch,
c          write(nout,'(" event = ",i9,",",(a))') nevt,
     &        ' process: lambdas <= 0'
          write(*,*) 'process: ',s,s1,s2,lambdas
          numout=numout+1
        endif
        return
      endif
      lambdat=(s-t1-t2)*(s-t1-t2)-4d0*t1*t2
      if(lambdat.gt.0d0)then
        lambdat=sqrt(lambdat)
      else
        switch=11
        if(numout.lt.maxout.and.lnoutmc.ge.5)then
          write(nout,100) nevt,cogen,cocha,switch,
c          write(nout,'(" event = ",i9,",",(a))') nevt,
     &        ' process: lambdat <= 0'
          numout=numout+1
        endif
        return
      endif
      tmin=s1+t1-((s+s1-s2)*(s+t1-t2)+lambdas*lambdat)/2d0/s
      tmax=s1+t1-((s+s1-s2)*(s+t1-t2)-lambdas*lambdat*cmax)/2d0/s
      if(tmax.le.tmin)then
        switch=12
        if(numout.lt.maxout.and.lnoutmc.ge.5)then
          write(nout,100) nevt,cogen,cocha,switch,
c          write(*,'(" event= ",i9,",",(a))'),nevt,
     &        ' process: tmin >= tmax'
          numout=numout+1
        endif
        return
      endif
      if(tmax.gt.0d0)then
        if(tmax.lt.techparam(2))then
          tmax=0d0
        else
          switch=13
          if(numout.lt.maxout.and.lnoutmc.ge.5)then
            write(nout,100) nevt,cogen,cocha,switch,
c            write(*,'(" event= ",i9,",",(a))'),nevt,
     &          ' process: tmax > 0'
            numout=numout+1
          endif
          return
        endif
      endif
      if(width.gt.0d0)then
        width2=width*width
        omin=datan((lmass2-tmax)/mass/width)
        omax=datan((lmass2-tmin)/mass/width)
        if(step.eq.1)then

c          write(*,*) 'process: use random=',random(1),random(2)

          phi=2d0*pi*random(1)
          omeco=omin+(omax-omin)*random(2)
          cost=((s+s1-s2)*(s+t1-t2)-2d0*s*(t1+s1)+2d0*lmass2*s
     *          -2d0*s*width*mass*dtan(omeco))/lambdas/lambdat
        elseif(step.eq.2)then
          denum=pi*(omax-omin)*((lmass2-t)**2+lmass2*width2)
          if(denum.ne.0d0)then
            g=g*2d0*lambdat*mass*width/denum
          else
            switch=14
            if(numout.lt.maxout.and.lnoutmc.ge.5)then
              write(nout,100) nevt,cogen,cocha,switch,
c              write(*,'(" event= ",i9,",",(a))'),nevt,
     &            ' process: 1/density = 0 (width>0)'
              numout=numout+1
            endif
            return
          endif
        endif
      else
        if(step.eq.1)then

c          write(*,*) 'process: use random=',random(1),random(2)

          phi=2d0*pi*random(1)
          denum=lambdas*lambdat
          cost=((s+s1-s2)*(s+t1-t2)-2d0*s*(t1+s1)
     *        -2d0*s*mapinv(random(2),lpower,-lmass2,-tmax,-tmin,switch)
     *        )/lambdas/lambdat
        elseif(step.eq.2)then
          denum=pi*jacobianinv(lpower,-lmass2,-t,-tmax,-tmin,switch)
          if(denum.ne.0d0)then
            g=g*2d0*lambdat/denum
          else
            switch=15
            if(numout.lt.maxout.and.lnoutmc.ge.5)then
              write(nout,100) nevt,cogen,cocha,switch,
c              write(*,'(" event= ",i9,",",(a))'),nevt,
     &            ' process: 1/density = 0 (width=0)'
            endif
            numout=numout+1
            return
          endif
        endif
      endif
      if(step.eq.1)then
        if(s.gt.0d0)then
          roots=sqrt(s)
        else
          switch=16
          if(numout.lt.maxout.and.lnoutmc.ge.5)then
            write(nout,100) nevt,cogen,cocha,switch,
c            write(*,'(" event= ",i9,",",(a))'),nevt,
     &          ' process: s<=0'
          endif
          numout=numout+1
          return
        endif
        p1(0)=(s+s1-s2)*0.5d0/roots
        p1(1)=0d0
        p1(2)=0d0
        p1(3)=lambdas*0.5d0/roots
        phi=dsign(phi,q1(3))
        call rotation(p1,phi,cost,switch)
        do i1=0,3
          q(i1)=q1(i1)+q2(i1)
          k1(i1)=q1(i1)
        enddo
        call boost(s,q,k1,1d0,switch)
        if(k1(1).eq.0d0)then
          phi=dsign(pi*0.5d0,k1(2))
        else
          phi=datan(k1(2)/k1(1))
        endif
        if(k1(1).lt.0d0)phi=phi+pi
        cost=k1(3)/sqrt(k1(1)**2+k1(2)**2+k1(3)**2)
        call rotation(p1,-phi,cost,switch)
        call boost(s,q,p1,-1d0,switch)
        do i1=0,3
          p2(i1)=q(i1)-p1(i1)
          qt(i1)=q1(i1)-p1(i1)
        enddo
        t=qt(0)**2-qt(1)**2-qt(2)**2-qt(3)**2
      endif
      end

***********************************************************************
      subroutine decay(random,q,p1,p2,g,s,s1,s2,step,switch)
***********************************************************************
*     mapping of  1->2 particle decay without s-channel propagator    *
*     input:                                                          *
*       random(2): random numbers  for angles                         *
*       q: incoming momentum                                          *
*       s: invariant mass of initial state                            *
*       s1,s2: invariant masses of outgoing momenta                   *
*     output:                                                         *
*       p1,p2: outgoing momenta           for step=1 and switch = 0   *
*       g: density                        for step=2 and switch = 0   *
*       switch > 0 for error                                          *
*---------------------------------------------------------------------*
*     by Markus Roth         last changed   14.11.06   Ansgar Denner  *
***********************************************************************
      implicit none

#include "mcparams.h"
#include "mccommon.h"

c local variable
      real*8 random(2),q(0:3),p1(0:3),p2(0:3)
      real*8 phi,cost,s,roots,lambda,denum,g,s1,s2
      integer i1,step,switch
      if(switch.gt.0)return
      lambda=(s-s1-s2)*(s-s1-s2)-4d0*s1*s2
      if(lambda.gt.0d0)then
        lambda=sqrt(lambda)
      else
        switch=17
        if(numout.lt.maxout.and.lnoutmc.ge.5)then
 100      format(/" event = ",i9,", gen = ",i4,", channel = ",i4,
     &        ", exit = ",i4,(a))
          write(nout,100) nevt,cogen,cocha,switch,
     &        ', decay: lambda <= 0'
c          write(nout,*)' decay: ',s,s1,s2,lambda
          write(nout,*)' decay: ',sqrt(s),sqrt(s1),sqrt(s2),lambda
          write(nout,*)' decay: ',sqrt(s)-sqrt(s1)-sqrt(s2)
          numout=numout+1
c          stop
        endif
        return
      endif
      if(step.eq.1)then
        if(s.gt.0d0)then
          roots=sqrt(s)
        else
          switch=18
          if(numout.lt.maxout.and.lnoutmc.ge.5)then
            write(nout,100) nevt,cogen,cocha,switch,
c            write(*,'(" event= ",i9,",",(a))'),nevt,
     &          ' decay: s <= 0'
            numout=numout+1
          endif
          return
        endif

c        write(*,*) 'decay: use random=',random(1),random(2)

        phi=2d0*pi*random(1)
        cost=2d0*random(2)-1d0
        p1(0)=(s+s1-s2)*0.5d0/roots
        p1(1)=0d0
        p1(2)=0d0
        p1(3)=lambda*0.5d0/roots
        call rotation(p1,phi,cost,switch)
        call boost(s,q,p1,-1d0,switch)
        do i1=0,3
          p2(i1)=q(i1)-p1(i1)
        enddo
      elseif(step.eq.2)then
        denum=pi*lambda
        if(denum.gt.0d0)then
          g=g*2d0*s/denum
        else
          switch=19
          if(numout.lt.maxout.and.lnoutmc.ge.5)then
            write(nout,100) nevt,cogen,cocha,switch,
c            write(*,'(" event= ",i9,",",(a))'),nevt,
     &          ' dec: 1/density <= 0'
            numout=numout+1
          endif
          return
        endif
      endif
      end

***********************************************************************
      function mapinv(random,lpower,lmass2,lxmin,lxmax,switch)
***********************************************************************
*     mapping for real propagators                                    *
*     input:                                                          *
*       random: random number                                         *
*       mass2:  mass squared                                          *
*       lpower: power of propagator, determines mapping               *
*       xmin,xmax: minimal and maximal values for invariant           *
*     output:                                                         *
*       smapinvh: invariant    for switch = 0                         *
*       switch > 0 for error                                          *
*     remarks                                                         *
*       xmin,xmax must be larger than mass2                           *
*             with allowed uncertainty determined by techparam(1)     *
*---------------------------------------------------------------------*
*     by Markus Roth         last changed   12.01.07   Ansgar Denner  *
***********************************************************************
      implicit none


#include "mcparams.h"
#include "mccommon.h"


      real*8 mapinv,random,lpower,lxmin,lxmax,lmass2,m2
      integer switch
c mctechparam
c      real*8 techparam(4)
c mcoutput
c      integer nout,numout,maxout,noutgen,nsteps
c      common/mctechparam/techparam
c      common/mcoutput/nout,numout,maxout,noutgen,nsteps

c      write(*,*) 'mapinv: use random=',random

      mapinv=0d0
      if(switch.gt.0)return
      m2=lmass2-techparam(1)
      if(lxmax-m2.lt.0d0.or.lxmin-m2.lt.0d0)then
        switch=19
        if(numout.lt.maxout.and.lnoutmc.ge.5)then
 100      format(/" event = ",i9,", gen = ",i4,", channel = ",i4,
     &        ", exit = ",i4,(a))
          write(nout,100) nevt,cogen,cocha,switch,
c          write(*,'(" event= ",i9,",",(a))'),nevt,
     &        ', h: xmin-m2, xmax-m2 < 0'
          numout=numout+1
        endif
        return
      endif
      if(lpower.eq.0d0)then
        mapinv=random*lxmax+(1d0-random)*lxmin
      elseif(lpower.eq.1d0)then
        mapinv=dexp(random*dlog(lxmax-m2)+(1d0-random)*dlog(lxmin-m2))
     &      +m2
      else
        mapinv=(random*(lxmax-m2)**(1d0-lpower)
     *    +(1d0-random)*(lxmin-m2)**(1d0-lpower))**(1d0/(1d0-lpower))+m2
      endif
      end

***********************************************************************
      function jacobianinv(lpower,lmass2,x,lxmin,lxmax,switch)
***********************************************************************
*     jacobian for mapping mapinv of real propagators                 *
*     input:                                                          *
*       lmass2:  mass squared                                         *
*       lpower: power of propagator, determines mapping               *
*       x:     value of invariant                                     *
*       lxmin,lxmax: minimal and maximal values for invariant         *
*     output:                                                         *
*       jacobianinv    for switch = 0                                 *
*       switch > 0 for error                                          *
*     remarks                                                         *
*       xmin,xmax must be larger than mass2                           *
*             with allowed uncertainty determined by techparam(1)     *
*---------------------------------------------------------------------*
*     by Markus Roth         last changed   07.01.07   Ansgar Denner  *
***********************************************************************
      implicit none
      real*8 jacobianinv,lpower,x,lxmin,lxmax,lmass2,m2
      integer switch

#include "mcparams.h"
#include "mccommon.h"

      jacobianinv=0d0
      if(switch.gt.0)return

c      write(*,*) 'jacobianinv: techparam= ',techparam

      m2=lmass2-techparam(1)

c      write(*,*) 'jacobianinv: m2= ',m2,xmax-m2,xmin-m2

      if(lxmax-m2.lt.0d0.or.lxmin-m2.lt.0d0.or.x-m2.lt.0d0)then
        switch=21
        if(numout.lt.maxout.and.lnoutmc.ge.5)then
 100      format(/" event = ",i9,", gen = ",i4,", channel = ",i4,
     &        ", exit = ",i4,(a))
          write(nout,100) nevt,cogen,cocha,switch,
     &        ', jacobianinv: xmin-m2, xmax-m2 < 0, x-m2 < 0'
          write(nout,*) ' jacobianinv: xmin,xmax,x,m2', lxmin,lxmax,x,m2
          write(nout,*) ' jacobianinv: xmin-m2,xmax-m2,x-m2',
     &        lxmin-m2,lxmax-m2,x-m2
          numout=numout+1
        endif
        return
      endif
      if(lpower.eq.0d0)then
        jacobianinv=lxmax-lxmin
      elseif(lpower.eq.1d0)then
        jacobianinv=(dlog(lxmax-m2)-dlog(lxmin-m2))*(x-m2)
      else
        jacobianinv=(((lxmax-m2)**(1d0-lpower)-(lxmin-m2)**(1d0-lpower))
     *    /(1d0-lpower))*(x-m2)**lpower
      endif
      end

***********************************************************************
      function mapx(random,lpower,lmass2,lxmin,lxmax,switch)
***********************************************************************
*     mapping for 1/(1-x)                                             *
*     input:                                                          *
*       random: random number                                         *
*       mass2:  mass squared                                          *
*       lpower: power of propagator, determines mapping               *
*       lxmin,lxmax: minimal and maximal values for invariant         *
*     output:                                                         *
*       mapx: invariant    for switch = 0                             *
*       switch > 0 for error                                          *
*     remarks                                                         *
*       xmin,xmax must be larger than mass2                           *
*---------------------------------------------------------------------*
*     by Markus Roth         last changed   12.01.07   Ansgar Denner  *
***********************************************************************
      implicit none

#include "mcparams.h"
#include "mccommon.h"

      real*8 mapx,random,lpower,lxmin,lxmax,lmass2,m2
      integer switch
c mctechparam
c      real*8 techparam(4)
c mcoutput
c      integer nout,numout,maxout,noutgen,nsteps
c      common/mctechparam/techparam
c      common/mcoutput/nout,numout,maxout,noutgen,nsteps

c       write(*,*) ' mapx in ',random,lpower,mass2,xmin,xmax,switch

c      write(*,*) 'mapx: use random=',random

      mapx=2d0
      if(switch.gt.0)return
c     m2=mass2-techparam(1)
      m2=lmass2
      if(lxmax-m2.lt.0d0.or.lxmin-m2.lt.0d0)then
        switch=40
        if(numout.lt.maxout.and.lnoutmc.ge.5)then
 100      format(/" event = ",i9,", gen = ",i4,", channel = ",i4,
     &        ", exit = ",i4,(a))
          write(nout,100) nevt,cogen,cocha,switch,
     &        ', mapx: xmin-m2, xmax-m2 < 0'
          numout=numout+1
        endif
        return
      endif
      if(lpower.eq.0d0)then
        mapx=random*lxmax+(1d0-random)*lxmin
      elseif(lpower.eq.1d0)then
        mapx=dexp(random*dlog(lxmax-m2)+(1d0-random)*dlog(lxmin-m2))+m2
      else
        mapx=(random*(lxmax-m2)**(1d0-lpower)
     *    +(1d0-random)*(lxmin-m2)**(1d0-lpower))**(1d0/(1d0-lpower))
     &      +m2
      endif
      end

***********************************************************************
      function jacobianx(lpower,lmass2,x,lxmin,lxmax,switch)
***********************************************************************
*     jacobian for mapping mapx for 1/(1-x)                           *
*     input:                                                          *
*       lmass2:  mass squared                                         *
*       lpower: power of propagator, determines mapping               *
*       x:     value of invariant                                     *
*       lxmin,lxmax: minimal and maximal values for invariant         *
*     output:                                                         *
*       jacobianx    for switch = 0                                   *
*       switch > 0 for error                                          *
*     remarks                                                         *
*       xmin,xmax must be larger than mass2                           *
*---------------------------------------------------------------------*
*     by Markus Roth         last changed   07.01.07   Ansgar Denner  *
***********************************************************************
      implicit none
      real*8 jacobianx,lpower,x,lxmin,lxmax,lmass2,m2
      integer switch

#include "mcparams.h"
#include "mccommon.h"

      jacobianx=0d0
      if(switch.gt.0)return

c      write(*,*) 'jacobianx: techparam= ',techparam

c     m2=lmass2-techparam(1)
      m2=lmass2

c      write(*,*) 'jacobianx: m2= ',m2,xmax-m2,xmin-m2

      if(lxmax-m2.lt.0d0.or.lxmin-m2.lt.0d0)then
        switch=21
        if(numout.lt.maxout.and.lnoutmc.ge.5)then
 100      format(/" event = ",i9,", gen = ",i4,", channel = ",i4,
     &        ", exit = ",i4,(a))
          write(nout,100) nevt,cogen,cocha,switch,
     &        ', jacobianx: xmin-m2, xmax-m2 < 0'
          write(nout,*)  'xmin-m2, xmax-m2 = ',lxmin-m2,lxmax-m2
          numout=numout+1
        endif
        return
      endif
      if(lpower.eq.0d0)then
        jacobianx=lxmax-lxmin
      elseif(lpower.eq.1d0)then
        jacobianx=(dlog(lxmax-m2)-dlog(lxmin-m2))*(x-m2)
      else
        jacobianx=(((lxmax-m2)**(1d0-lpower)-(lxmin-m2)**(1d0-lpower))
     *    /(1d0-lpower))*(x-m2)**lpower
      endif
      end

***********************************************************************
      subroutine rotation(p,phi,cost,switch)
***********************************************************************
*     rotation (C15) of Markus Roth's PhD thesis                      *
*     rotates momentum (0,0,1) to                                     *
*      (sin(theta) cos(phi),-sin(theta) sin(phi),cos(theta))          *
*---------------------------------------------------------------------*
*     by Markus Roth         last changed   14.11.06   Ansgar Denner  *
***********************************************************************
      implicit none

#include "mcparams.h"
#include "mccommon.h"

c local variable
      real*8 p(0:3),phi,cost,sint,cosp,sinp,px,py,pz
      integer switch
c mcoutput
c      integer nout,numout,maxout,noutgen,nsteps
c      common/mcoutput/nout,numout,maxout,noutgen,nsteps

      if(switch.gt.0)return
      cosp=dcos(phi)
      sinp=dsin(phi)
      if(dabs(cost).gt.1d0)then
        switch=22
        if(numout.lt.maxout.and.lnoutmc.ge.5)then
 100      format(/" event = ",i9,", gen = ",i4,", channel = ",i4,
     &        ", exit = ",i4,(a))
          write(nout,100) nevt,cogen,cocha,switch,
     &        ', rotation: |cost| > 1'
          numout=numout+1
        endif
        return
      endif
      sint=sqrt((1d0-cost)*(1d0+cost))
      px=p(1)
      py=p(2)
      pz=p(3)
      p(1)=(pz*sint+px*cost)*cosp+py*sinp
      p(2)=-(pz*sint+px*cost)*sinp+py*cosp
      p(3)=pz*cost-px*sint
      end

***********************************************************************
      subroutine boost(q2,q,p,dir,switch)
***********************************************************************
*     boost                                                           *
*     dir = +1: boosts p to rest frame of q                           *
*     dir = -1: boosts p from rest frame of q to frame of q           *
*     input                                                           *
*     q2        square of q                                           *
*     q         momentum that fixes boost                             *
*     p         input momentum                                        *
*     input                                                           *
*     p         boosted momentum                                      *
*---------------------------------------------------------------------*
*     by Markus Roth            last changed  14.11.06 Ansgar Denner  *
***********************************************************************
      implicit none

#include "mcparams.h"
#include "mccommon.h"

c local variable
      real*8 p(0:3),q(0:3),dir,m,q2,m2,bx,by,bz,gamma
      real*8 p0,px,py,pz,bp,a
      integer switch

      if(switch.gt.0)return
      if(q2.eq.0d0) then
        m2=q(0)**2-q(1)**2-q(2)**2-q(3)**2
      elseif(q2.gt.0d0)then
        m2=q2
        if(dabs(q(0)**2-q(1)**2-q(2)**2-q(3)**2-q2).gt.1d-5*q(0)**2)
     &      then
          write(*,*) 'boost: q2 and m2 do not agree:',q2,
     &       q(0)**2-q(1)**2-q(2)**2-q(3)**2
        endif
      else
        write(*,*) 'boost: 0 > q2 =',q2
        stop
      endif
      if(m2.le.0d0)then
        switch=23
        if(numout.lt.maxout.and.lnoutmc.ge.5)then
 100      format(/" event = ",i9,", gen = ",i4,", channel = ",i4,
     &        ", exit = ",i4,(a))
          write(nout,100) nevt,cocha,switch
          write(nout,*)
     &        ', boost: m2 <= 0',m2
          numout=numout+1
          write(nout,*)' boost: q',q
        endif
        return
      endif

      m=sqrt(m2)
      bx=-q(1)/m*dir
      by=-q(2)/m*dir
      bz=-q(3)/m*dir
      gamma=q(0)/m
      a=1d0/(1d0+gamma)
      p0=p(0)
      px=p(1)
      py=p(2)
      pz=p(3)
      bp=bx*px+by*py+bz*pz
      p(0)=gamma*p0+bp
      p(1)=px+bx*p0+a*bp*bx
      p(2)=py+by*p0+a*bp*by
      p(3)=pz+bz*p0+a*bp*bz
      end

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                 c
c     mapping from (n+1)- into n-particle phase space             c
c                                                                 c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine mapin(kt,g,em,sp,co,lnext,switch)
      implicit none

#include "mcparams.h"
#include "mccommon.h"

c local variables
      real*8 k(maxe,0:3),kt(maxe,0:3),g,denum,jacobianx
      real*8 x,z,y,v,kk(7,7),ck(0:3),ckt(0:3),st
      integer i1,i2,j,em,sp,co,lnext,switch

c to avoid compiler warning
      data  denum/0d0/, x/0d0/

      if(switch.gt.0)return
      if(em.eq.0.or.sp.eq.0.or.co.eq.0)then
        g=1d0
        return
      endif
      do i1=1,lnext
      do i2=0,3
        k(i1,i2)=kt(i1,i2)
      enddo
      enddo
      kk(em,sp)=k(em,0)*k(sp,0)-k(em,1)*k(sp,1)-k(em,2)*k(sp,2)
     *  -k(em,3)*k(sp,3)
      kk(em,co)=k(em,0)*k(co,0)-k(em,1)*k(co,1)-k(em,2)*k(co,2)
     *  -k(em,3)*k(co,3)
      kk(sp,co)=k(sp,0)*k(co,0)-k(sp,1)*k(co,1)-k(sp,2)*k(co,2)
     *  -k(sp,3)*k(co,3)
c final-state emitter with final-state spectator
      if(em.ge.3.and.sp.ge.3)then
        x=1d0
        z=kk(em,sp)/(kk(em,sp)+kk(sp,co))
        y=kk(em,co)/(kk(em,sp)+kk(em,co)+kk(sp,co))
        do i1=0,3
          kt(em,i1)=k(em,i1)+k(co,i1)-y/(1d0-y)*k(sp,i1)
          kt(sp,i1)=1d0/(1d0-y)*k(sp,i1)
          kt(co,i1)=0d0
        enddo
        denum=pi*(1d0-y)*jacobianx(powerexmap,0d0,1d0-y,0d0,1d0,switch)
     *    *jacobianx(powerexmap,0d0,1d0-z,0d0,1d0,switch)
c final-state emitter with initial-state spectator
      elseif(em.ge.3.and.sp.le.2)then
        x=(kk(em,sp)+kk(sp,co)-kk(em,co))/(kk(em,sp)+kk(sp,co))
        z=kk(em,sp)/(kk(em,sp)+kk(sp,co))
        do i1=0,3
          kt(em,i1)=k(em,i1)+k(co,i1)-(1d0-x)*k(sp,i1)
          kt(sp,i1)=x*k(sp,i1)
          kt(co,i1)=0d0
        enddo
c CHANGED 17.11.06
c        denum=pi*jacobianinv(powerexmap,0d0,1d0-x,0d0,1d0,switch)
        denum=pi*jacobianx(powerexmap,0d0,1d0-x,0d0,1d0-xminp,switch)
     *    *jacobianx(powerexmap,0d0,1d0-z,0d0,1d0,switch)
c initial-state emitter with final-state spectator
      elseif(em.le.2.and.sp.ge.3)then
        x=(kk(em,sp)+kk(em,co)-kk(sp,co))/(kk(em,sp)+kk(em,co))
        z=kk(em,sp)/(kk(em,sp)+kk(em,co))
        do i1=0,3
          kt(sp,i1)=k(sp,i1)+k(co,i1)-(1d0-x)*k(em,i1)
          kt(em,i1)=x*k(em,i1)
          kt(co,i1)=0d0
        enddo
c CHANGED 17.11.06
c        denum=pi*jacobianinv(powerexmap,0d0,1d0-x,0d0,1d0,switch)
        denum=pi*jacobianx(powerexmap,0d0,1d0-x,0d0,1d0-xminp,switch)
     *    *jacobianx(powerexmap,0d0,1d0-z,0d0,1d0,switch)
c initial-state emitter with initial-state spectator
      elseif(em.le.2.and.sp.le.2)then
        x=(kk(em,sp)-kk(em,co)-kk(sp,co))/kk(em,sp)
        v=kk(em,co)/kk(em,sp)
        do i1=0,3
          ck(i1)=k(em,i1)+k(sp,i1)-k(co,i1)
          ckt(i1)=x*k(em,i1)+k(sp,i1)
        enddo
        do j=3,6
          kk(em,j)=k(em,0)*k(j,0)-k(em,1)*k(j,1)-k(em,2)*k(j,2)
     *      -k(em,3)*k(j,3)
          kk(sp,j)=k(sp,0)*k(j,0)-k(sp,1)*k(j,1)-k(sp,2)*k(j,2)
     *      -k(sp,3)*k(j,3)
          kk(co,j)=k(co,0)*k(j,0)-k(co,1)*k(j,1)-k(co,2)*k(j,2)
     *      -k(co,3)*k(j,3)
          do i1=0,3
            kt(j,i1)=k(j,i1)
     *        -((1d0+x)*kk(em,j)+2d0*kk(sp,j)-kk(co,j))
     *          /(4d0*x+v-x*v)/kk(em,sp)*(ck(i1)+ckt(i1))
     *        +(kk(em,j)+kk(sp,j)-kk(co,j))/x/kk(em,sp)*ckt(i1)
          enddo
        enddo
        do i1=0,3
          kt(em,i1)=x*k(em,i1)
          kt(sp,i1)=k(sp,i1)
          kt(co,i1)=0d0
        enddo
c CHANGED 17.11.06
c        denum=pi*(1d0-x)*jacobianinv(powerexmap,0d0,1d0-x,0d0,1d0,switch)
        denum=pi*(1d0-x)*jacobianx(powerexmap,0d0,1d0-x,0d0,
     &      1d0-xminp,switch)
      endif
      st=kt(em,0)*kt(sp,0)-kt(em,1)*kt(sp,1)-kt(em,2)*kt(sp,2)
     *  -kt(em,3)*kt(sp,3)
      if(denum.gt.0d0.and.st.ne.0d0)then
        g=x/denum/st
      else
        switch=24
        if(numout.lt.maxout.and.lnoutmc.ge.5)then
 100      format(/" event = ",i9,", gen = ",i4,", channel = ",i4,
     &        ", exit = ",i4,(a))
          write(nout,100) nevt,cogen,cocha,switch,
     &        ', mapin: 1/g <= 0'
          numout=numout+1
        endif
        return
      endif
      end

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                 c
c     mapping from n- into (n+1)-particle phase space             c
c                                                                 c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine mapout(random,k,x,g,em,sp,co,lnext,switch)
***********************************************************************
*     maps momenta of n-particle phase space to momenta of            *
*    (n+1)-particle phase space                                       *
*     input:                                                          *
*       random:    random numbers                                     *
*       k:         external particle momenta of n-particle phase space*
*       em:        emitter particle                                   *
*       sp:        spectator particle                                 *
*       co:        collinear (emitted) particle                       *
*       lnext:     number of external particles                       *
*     output:                                                         *
*       k:         external part. momenta of n+1 particle phase space *
*       switch > 0 for error                                          *
*     adapted from COFFERGG by Markus Roth                            *
*---------------------------------------------------------------------*
*     by Markus Roth         last changed   06.03.07   Ansgar Denner  *
***********************************************************************
      implicit none

#include "mcparams.h"
#include "mccommon.h"

c local variables
      real*8 random(3),k(maxe,0:3),kt(maxe,0:3),g,m2,m
      real*8 kem(0:3),ksp(0:3),ktr(0:3),ck(0:3),ckt(0:3)
      real*8 ksum(0:3),y,mapx,z,f,v,c,x,kk(maxe,maxe),omz,omy
      integer i1,i2,j,em,sp,co,lnext,switch

c      write(*,*) 'mapout: use random=',random(1),random(2),random(3)

      if(switch.gt.0)return
      if(em.eq.0.and.sp.eq.0.and.co.eq.0)return

c mapping from n- into (n+1)-particle phase space
      do i1=0,3
        kem(i1)=k(em,i1)
        ksp(i1)=k(sp,i1)
        ksum(i1)=kem(i1)+ksp(i1)
      enddo
      m2=ksum(0)**2-ksum(1)**2-ksum(2)**2-ksum(3)**2
      if(m2.le.0d0)then
        switch=25
        if(numout.lt.maxout.and.lnoutmc.ge.5)then
 100      format(/" event = ",i9,", gen = ",i4,", channel = ",i4,
     &        ", exit = ",i4,(a))
          write(nout,100) nevt,cogen,cocha,switch,
     &        ', mapout: m2 <= 0'
          numout=numout+1
        endif
        return
      else
        m=sqrt(m2)
      endif
      if(em.ge.3.and.sp.ge.3)then
c final-state emitter with final-state spectator
        y=1d0-mapx(random(1),powerexmap,0d0,0d0,1d0,switch)
        omz=mapx(random(2),powerexmap,0d0,0d0,1d0,switch)
        z=1d0-omz
        f=sqrt(y*z/omz)
        ktr(0)=-f*f*m
        ktr(1)=f*m*dcos(2d0*pi*random(3))
        ktr(2)=f*m*dsin(2d0*pi*random(3))
        ktr(3)=-f*f*m
        call transverse(ksum,ksp,ktr,switch)
        do i1=0,3
          k(em,i1)=y*(1d0+z)*ksp(i1)+z*kem(i1)+omz*ktr(i1)
          k(co,i1)=omz*(kem(i1)-ktr(i1))-y*z*ksp(i1)
          k(sp,i1)=(1d0-y)*ksp(i1)
        enddo
c        write(*,*) 'mapout ff',em,co,sp,k
c final-state emitter with initial-state spectator
      elseif(em.ge.3.and.sp.le.2)then
        omz=mapx(random(2),powerexmap,0d0,0d0,1d0,switch)
        z=1d0-omz
        f=sqrt(z*(1d0-x)/(omz*x))
        ktr(0)=-f*f*m
        ktr(1)=f*m*dcos(2d0*pi*random(3))
        ktr(2)=f*m*dsin(2d0*pi*random(3))
        ktr(3)=-f*f*m
        call transverse(ksum,ksp,ktr,switch)
        do i1=0,3
          k(co,i1)=omz*(kem(i1)-ktr(i1))-(1d0-x)/x*z*ksp(i1)
          k(em,i1)=(1d0-x)/x*(1d0+z)*ksp(i1)+z*kem(i1)
     *      +omz*ktr(i1)
          k(sp,i1)=ksp(i1)/x
        enddo
c        write(*,*) 'mapout fi',em,co,sp,k
c initial-state emitter with final-state spectator
      elseif(em.le.2.and.sp.ge.3)then
        omz=mapx(random(2),powerexmap,0d0,0d0,1d0,switch)
        z=1d0-omz
        f=sqrt(omz*(1d0-x)/(z*x))
        ktr(0)=-f*f*m
        ktr(1)=f*m*dcos(2d0*pi*random(3))
        ktr(2)=f*m*dsin(2d0*pi*random(3))
        ktr(3)=-f*f*m
        call transverse(ksum,kem,ktr,switch)
        do i1=0,3
          k(co,i1)=(1d0-x)/x*(2d0-z)*kem(i1)+omz*ksp(i1)
     *      +z*ktr(i1)
          k(sp,i1)=z*(ksp(i1)-ktr(i1))-(1d0-x)/x*omz*kem(i1)
          k(em,i1)=kem(i1)/x
        enddo
c        write(*,*) 'mapout if',em,co,sp,k
c initial-state emitter with initial-state spectator
      elseif(em.le.2.and.sp.le.2)then
        do i1=3,lnext
        do i2=0,3
          kt(i1,i2)=k(i1,i2)
        enddo
        enddo
        v=random(2)*(1d0-x)
        f=sqrt(v*(1d0-x-v)/x)
        ktr(0)=0d0
        ktr(1)=f*m*dcos(2d0*pi*random(3))
        ktr(2)=f*m*dsin(2d0*pi*random(3))
        ktr(3)=0d0
        call transverse(ksum,kem,ktr,switch)
        do i1=0,3
          ck(i1)=(x+v)/x*kem(i1)+(1d0-v)*ksp(i1)-ktr(i1)
          ckt(i1)=kem(i1)+ksp(i1)
        enddo
        do i1=0,3
          k(co,i1)=v*ksp(i1)+(1d0-x-v)/x*kem(i1)+ktr(i1)
          k(em,i1)=kem(i1)/x
        enddo
        do j=3,lnext
        if(j.ne.co)then
          c=ktr(0)*kt(j,0)-ktr(1)*kt(j,1)-ktr(2)*kt(j,2)
     *      -ktr(3)*kt(j,3)
          kk(em,sp)=kem(0)*ksp(0)-kem(1)*ksp(1)-kem(2)*ksp(2)
     *      -kem(3)*ksp(3)
          kk(em,j)=kt(j,0)*kem(0)-kt(j,1)*kem(1)-kt(j,2)*kem(2)
     *      -kt(j,3)*kem(3)
          kk(sp,j)=kt(j,0)*ksp(0)-kt(j,1)*ksp(1)-kt(j,2)*ksp(2)
     *      -kt(j,3)*ksp(3)
          do i1=0,3
            k(j,i1)=kt(j,i1)
     *        -((2d0*x+v)*kk(em,j)+(2d0-v)*x*kk(sp,j)-x*c)
     *         /((4d0*x+v-x*v)*kk(em,sp))*(ck(i1)+ckt(i1))
     *        +(kk(em,j)+kk(sp,j))/kk(em,sp)*ck(i1)
          enddo
        endif
        enddo
c        write(*,*) 'mapout ii',em,co,sp,k
      endif
      end

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                 c
c     determination of transverse momenta                         c
c                                                                 c
c     written by Markus Roth                                      c
c                                                                 c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine transverse(ksum,k,ktr,switch)
      implicit none

#include "mcparams.h"
#include "mccommon.h"

c local variables
      real*8 k(0:3),ktr(0:3),ksum(0:3),p(0:3),pvec,phi,cost
      integer i1,switch
c output
c      integer nout,numout,maxout
c      common/output/nout,numout,maxout

      if(switch.gt.0)return
c      pi=4d0*datan(1d0)
      do i1=0,3
        p(i1)=k(i1)
      enddo
      call boost(0d0,ksum,p,1d0,switch)
      if(p(1).eq.0d0)then
        phi=dsign(pi*0.5d0,p(2))
      else
        phi=datan(p(2)/p(1))
      endif
      if(p(1).lt.0d0)phi=phi+pi
      if(p(1).gt.0d0.and.p(2).lt.0d0)phi=phi+2d0*pi
      pvec=sqrt(p(1)**2+p(2)**2+p(3)**2)
      if(pvec.eq.0d0)then
        switch=26
        if(numout.lt.maxout.and.lnoutmc.ge.5)then
 100      format(/" event = ",i9,", gen = ",i4,", channel = ",i4,
     &        ", exit = ",i4,(a))
          write(nout,100) nevt,cogen,cocha,switch,
     &        ', transverse: pvec = 0'
          numout=numout+1
        endif
        return
      endif
      cost=p(3)/pvec
      call rotation(ktr,-phi,cost,switch)
      call boost(0d0,ksum,ktr,-1d0,switch)

      end

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                 c
c     3-particle vertex                                           c
c                                                                 c
c     written by Markus Roth                                      c
c                                                                 c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      function vertex(i1,i2,i3,schannel)
      implicit none

#include "mcparams.h"
#include "mccommon.h"

c local variables
      integer i1,i2,i3
      logical vertex,vertexg,schannel

      vertex=.false.
      if(gname(i1).eq.'   ')return
      if(gname(i2).eq.'   ')return
      if(gname(i3).eq.'   ')return
c mcfamily conservation
      if(mcfamily(i1,1)+mcfamily(i2,1)+mcfamily(i3,1).ne.0)return
      if(mcfamily(i1,2)+mcfamily(i2,2)+mcfamily(i3,2).ne.0)return
      if(mcfamily(i1,3)+mcfamily(i2,3)+mcfamily(i3,3).ne.0)return
      if(mcfamily(i1,4)+mcfamily(i2,4)+mcfamily(i3,4).ne.0)return
      if(mcfamily(i1,5)+mcfamily(i2,5)+mcfamily(i3,5).ne.0)return
      if(mcfamily(i1,6)+mcfamily(i2,6)+mcfamily(i3,6).ne.0)return
      vertex=.true.
c testing vertex
      if(vertexg(i1,i2,i3,schannel))return
      if(vertexg(i1,i3,i2,schannel))return
      if(vertexg(i2,i1,i3,schannel))return
      if(vertexg(i2,i3,i1,schannel))return
      if(vertexg(i3,i1,i2,schannel))return
      if(vertexg(i3,i2,i1,schannel))return
      vertex=.false.
      end


ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                 c
c     identifying particles (usually not used)                    c
c                                                                 c
c     written by Markus Roth                                      c
c                                                                 c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      function pid(binary)
      implicit none
c local variables
      integer pid,binary,i1,i2,i3,nexternal
      i3=1
      nexternal=30
      pid=0
      do i1=nexternal,1,-1
        i2=binary/2**(i1-1)
        if((i2/2)*2.ne.i2)then
          pid=pid+i1*i3
          i3=10*i3
        endif
      enddo
      end

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                 c
c     adaptive optimization                                       c
c                                                                 c
c     written by Markus Roth                                      c
c                                                                 c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
***********************************************************************
      subroutine optimization(lweight,g,gsum,n,generator,lnchannel)
***********************************************************************
*     adaptitive weight optimization                                  *
*     according to Kleiss, Pittau, Comput. Phys. Commun 75 (1994 141  *
*     taken from COFFERGG by Markus Roth                              *
*     input: lweight    weight  of event                              *
*            g(maxch)   corresponding densities of channels           *
*            gsum       corresponding total density                   *
*            n          number of event                               *
*            generator  number of generator                           *
*            lnchannel  number of channels                            *
*---------------------------------------------------------------------*
*     2005      Markus Roth     last changed  00.00.00  Ansgar Denner *
***********************************************************************
      implicit none
c
#include "mcparams.h"
#include "mccommon.h"

c local variables
      real*8 lweight,g(maxch),gsum
      integer n,lnchannel,generator
      real*8 dis(0:maxg,maxo),mindis(0:maxg),ldis
      integer i1,i2

      integer iNumProcs,iMyRank
      common/mpistuff/iNumProcs,iMyRank
      integer do_optimize
      common/do_optimize_common/do_optimize

      data ldis /0d0/, mindis /0d0,maxg*0d0/

      if(nopt(opt(generator),generator).ne.0)then
        if(lweight.ne.0d0.and.gsum.ne.0d0)then
          do i1=1,lnchannel
            if(g(i1)*gsum.lt.0d0)then
              if(numout.lt.maxout.and.lnoutmc.ge.5)then
                write(nout,'(" event= ",i9,",",(a))') nevt,
     &              ' optimization: gi*gsum <= 0'
                numout=numout+1
              endif
              return
            else
              wi(i1,generator)=wi(i1,generator)
c changed 5.05.06 AD
c     *          +sqrt(g(i1)*lweight*lweight/gsum)
     *          + g(i1)*lweight*lweight/gsum
            endif
          enddo
        endif
#ifdef mpiuse
        if((iMyRank.EQ.0 .AND. do_optimize.EQ.1) .OR.
     &      iNumProcs.EQ.1) Then
#endif
c --> added 5.05.06 AD
        if(n.ge.nopt(opt(generator),generator).and.nevents.gt.2*n)then
          ldis=0d0
          do i1=1,lnchannel
            do i2=1,lnchannel
              if( wi(i2,generator).ne.0d0) then
                ldis = max(ldis,
     &              abs(wi(i1,generator)/wi(i2,generator)))
              endif
            enddo
          enddo
          dis(generator,opt(generator)) = ldis
c          write(*,*) 'optimization: wi:',n,
c     &        (wi(i1,generator),i1=1,lnchannel),ldis
          if (opt(generator).eq.1.or.
     &        ldis.lt.mindis(generator)) then
            mindis(generator) = ldis
c --<
            do i1=1,lnchannel
              alphaopt(i1,opt(generator)+1,generator)=
     *            alphaopt(i1,opt(generator),generator)
     *            *sqrt(wi(i1,generator))
              betaopt(i1,generator)=betaopt(i1-1,generator)
     *            +alphaopt(i1,opt(generator)+1,generator)
              wi(i1,generator)=0d0
            enddo
c --> added 5.05.06 AD
          else
            do i1=1,lnchannel
              alphaopt(i1,opt(generator)+1,generator)=
     *            alphaopt(i1,opt(generator),generator)
              betaopt(i1,generator)=betaopt(i1-1,generator)
     *            +alphaopt(i1,opt(generator)+1,generator)
              wi(i1,generator)=0d0
            enddo
          endif
c          write(*,*) 'optimization: generator:',generator
c          write(*,*) 'optimization: alpha:',
c     &        (alphaopt(i1,opt(generator)+1,generator),i1=1,lnchannel)
c          write(*,*) 'optimization: beta:',
c     &        (betaopt(i1,generator),i1=1,lnchannel)

c --<
          opt(generator)=opt(generator)+1
          do i1=1,lnchannel
            if(betaopt(lnchannel,generator).ne.0d0)then
              alphaopt(i1,opt(generator),generator)=alphamin
     *            /dble(lnchannel)
     *            +alphaopt(i1,opt(generator),generator)*(1d0-alphamin)
     *            /betaopt(lnchannel,generator)
              betaopt(i1,generator)=alphamin*dble(i1)/dble(lnchannel)
     *            +betaopt(i1,generator)*(1d0-alphamin)
     *            /betaopt(lnchannel,generator)
            else
              alphaopt(i1,opt(generator),generator)=1d0/dble(lnchannel)
              betaopt(i1,generator)=dble(i1)/dble(lnchannel)
            endif
          enddo
c          write(*,*) 'optimization: generator:',generator
c          write(*,*) 'optimization: alpha:',
c     &        (alphaopt(i1,opt(generator),generator),i1=1,lnchannel)
c          write(*,*) 'optimization: beta:',
c     &        (betaopt(i1,generator),i1=1,lnchannel)
        else if (n.ge.nevents-1) then
          ldis=0d0
          do i1=1,lnchannel
            do i2=1,lnchannel
              if( wi(i2,generator).ne.0d0) then
                ldis = max(ldis,
     &              abs(wi(i1,generator)/wi(i2,generator)))
              endif
            enddo
          enddo
            dis(generator,opt(generator)) = ldis
c            write(*,*) 'optimization: wi:',generator,n,
c     &          (wi(i1,generator),i1=1,lnchannel),ldis
          endif
#ifdef mpiuse
        endif
#endif
        endif
      end


ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                 c
c     pseudo-random number generator                              c
c                                                                 c
c     written by Markus Roth                                      c
c                                                                 c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine rans(random)
      implicit none
c local variables
      real*8 s1,s2,s3,random
      integer init
      save init,s1,s2,s3
      data init/0/,s1/0d0/,s2/0d0/,s3/0d0/
      if (init.eq.0) then
        init=1
        s1=sqrt(2d0)-1d0
        s2=sqrt(3d0)-1d0
        s3=sqrt(5d0)-2d0
      endif
      s1=dmod(s1+s2+s3,1d0)
      s2=dmod(s1+s2+s3,1d0)
      s3=dmod(s1+s2+s3,1d0)
      random=s1
c      write(*,*) 'rans ',s1,s2,s3
      end
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                 c
c            rcarry number generator                              c
c                                                                 c
c     written by Andreas Kaiser                                   c
c                                                                 c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

      subroutine ransrcarry(realnumber)
      implicit none
      integer i1,zahl
      real*8 realnumber

      common/inzahl/zahl

      if(zahl.gt.134) then
         zahl=1
         do i1=1,93
            call rcarry(realnumber)
         enddo
      else
         zahl=zahl+1
      endif
      call rcarry(realnumber)
      end

      subroutine rcarry(realnumber)
      implicit none
      integer s,r,B
c      parameter (s=10,r=24,B=16777216) ! note that r>s   B=2^24
      parameter (s=10,r=24,B=1073741824) ! note that r>s  B=2^30

      real*8 realnumber
      integer i,z,xi
      integer x(0:r),carry,initsteps

      common/randomseed/x,carry,initsteps

      xi=x(s)-x(r)-carry
      if(xi.ge.0) then
         z=xi
         carry=0
      else
         z=xi+B
         carry=1
      endif
c do the shift
      do i=r,1,-1
         x(i)=x(i-1)
      enddo
      x(0)=z
      realnumber=dble(x(0))/B
      end

      subroutine initcarry
      implicit none
      integer x(0:24),carry,i1,zahl,initsteps
      double precision ran
      common/randomseed/x,carry,initsteps
      common/inzahl/zahl

      integer nout,numout,maxout,noutgen,nsteps,lnoutmc
      common/mcoutput/nout,numout,maxout,noutgen,nsteps,lnoutmc

       do i1=0,24
          x(i1)=i1
       enddo
       carry=0
       zahl=999999

c christoph's choice  a
c      initsteps=100000
c testchoice 1        b
c       initsteps=0
c testchoice 2        c
c       initsteps=1000
c testchoice 3        d
c      initsteps=200000

       do i1=1,initsteps
          call ransrcarry(ran)
       enddo

       if(lnoutmc.ge.5) then
         write(nout,'(a15,i20)') ' initsteps =',initsteps
       endif

      end
